import java.io.BufferedReader;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URI;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.hadoop.conf.Configurable;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.RawComparator;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.WritableComparable;
import org.apache.hadoop.io.WritableComparator;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.KeyValueTextInputFormat;
import org.apache.hadoop.mapred.lib.InputSampler;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Partitioner;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.Reducer.Context;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.TextInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;






//extends Configured implements Tool 
public class CloudJoin04Ext extends Configured implements Tool 
{
	private static final Logger log = Logger.getLogger(CloudJoin04Ext.class);
	private static final double errorAdjsmt = 0.000000000000005;
	private static final int constSmallNum = 100;
	private static final int numFilesToGen = 1;
	private static final int size = 16;


 	public static class VectorElem9D implements WritableComparable<VectorElem9D>
	{
		private long		partitionID		= -1;
		private long		windowID		= -1;
		private long		key;
		private double 	e1;
		private double 	e2;
		private double 	e3;
		private double 	e4;
		private double 	e5;
		private double 	e6;
		private double 	e7;
		private double 	e8;
		private double 	e9;
		private long		prevPartition 	= -1;
		
		VectorElem9D()
		{
			
		}
		
		VectorElem9D(long key, double e1,double e2, double e3, double e4, double e5, double e6, double e7, double e8, double e9)
		{
			this.key = key;
			this.e1 = e1;
			this.e2 = e2;
			this.e3 = e3;
			this.e4 = e4;
			this.e5 = e5;
			this.e6 = e6;
			this.e7 = e7;
			this.e8 = e8;
			this.e9 = e9;
		}
		
		//this could be an abstract method in an abstract class in, 
		//take a text type and populate the values based off the input
		VectorElem9D(Text text)
		{
			StringTokenizer line = new StringTokenizer(text.toString());
			this.key = Long.parseLong(line.nextToken());
			this.e1 = Double.parseDouble(line.nextToken());
			this.e2 = Double.parseDouble(line.nextToken());
			this.e3 = Double.parseDouble(line.nextToken());
			this.e4 = Double.parseDouble(line.nextToken());
			this.e5 = Double.parseDouble(line.nextToken());
			this.e6 = Double.parseDouble(line.nextToken());
			this.e7 = Double.parseDouble(line.nextToken());
			this.e8 = Double.parseDouble(line.nextToken());
			this.e9 = Double.parseDouble(line.nextToken());
			if(line.hasMoreElements())
				this.prevPartition = Long.parseLong(line.nextToken());
		}
		
		VectorElem9D(String text) 
		{
			StringTokenizer line = new StringTokenizer(text);
			this.key = Long.parseLong(line.nextToken());
			this.e1 = Double.parseDouble(line.nextToken());
			this.e2 = Double.parseDouble(line.nextToken());
			this.e3 = Double.parseDouble(line.nextToken());
			this.e4 = Double.parseDouble(line.nextToken());
			this.e5 = Double.parseDouble(line.nextToken());
			this.e6 = Double.parseDouble(line.nextToken());
			this.e7 = Double.parseDouble(line.nextToken());
			this.e8 = Double.parseDouble(line.nextToken());
			this.e9 = Double.parseDouble(line.nextToken());
			if(line.hasMoreElements())
				this.prevPartition = Long.parseLong(line.nextToken());
		}
		
		public int getSizeInBytes()
		{
			return (( 	Long.SIZE 	+		//partitionID
						Long.SIZE 	+		//windowID
						Long.SIZE 	+		//key
						((Double.SIZE) * 9)	+		//elems
						Long.SIZE 			//prevPartition
					)/8);
			
			
		}
		
		public long getKey()
		{
			return key;
		}
		
		public void setPartitionID(long partitionID) {
			this.partitionID = partitionID;
		}

		public long getPartitionID() {
			return partitionID;
		}

		public long getWindowID() {
			return windowID;
		}

		public void setWindowID(long windowID) {
			this.windowID = windowID;
		}
		
		public long getPrevPartition() {
			return prevPartition;
		}

		public void setPrevPartition(long prevPartition) {
			this.prevPartition = prevPartition;
		}
		
		public double getE1() {
			return e1;
		}

		public void setE1(double e1) {
			this.e1 = e1;
		}

		public double getE2() {
			return e2;
		}

		public void setE2(double e2) {
			this.e2 = e2;
		}

		public double getE3() {
			return e3;
		}

		public void setE3(double e3) {
			this.e3 = e3;
		}

		public double getE4() {
			return e4;
		}

		public void setE4(double e4) {
			this.e4 = e4;
		}

		public double getE5() {
			return e5;
		}

		public void setE5(double e5) {
			this.e5 = e5;
		}

		public double getE6() {
			return e6;
		}

		public void setE6(double e6) {
			this.e6 = e6;
		}

		public double getE7() {
			return e7;
		}

		public void setE7(double e7) {
			this.e7 = e7;
		}

		public double getE8() {
			return e8;
		}

		public void setE8(double e8) {
			this.e8 = e8;
		}

		public double getE9() {
			return e9;
		}

		public void setE9(double e9) {
			this.e9 = e9;
		}

		//would need to be implemented in an abstract class
		public double getDistanceBetween(VectorElem9D o)
		{
			//double deltaX = this.x - o.getX();
			//double deltaY = this.y - o.getY();
			
			return Math.sqrt(
					Math.pow((this.e1 - o.getE1()), 2) + 
					Math.pow((this.e2 - o.getE2()), 2) + 
					Math.pow((this.e3 - o.getE3()), 2) + 
					Math.pow((this.e4 - o.getE4()), 2) + 
					Math.pow((this.e5 - o.getE5()), 2) + 
					Math.pow((this.e6 - o.getE6()), 2) + 
					Math.pow((this.e7 - o.getE7()), 2) + 
					Math.pow((this.e8 - o.getE8()), 2) + 
					Math.pow((this.e9 - o.getE9()), 2)
					) ;
		}

		@Override
		public void readFields(DataInput in) throws IOException 
		{
			this.partitionID 	= in.readLong();
			this.windowID 		= in.readLong();
			this.key 			= in.readLong();
			this.e1 			= in.readDouble();
			this.e2 			= in.readDouble();
			this.e3 			= in.readDouble();
			this.e4 			= in.readDouble();
			this.e5 			= in.readDouble();
			this.e6 			= in.readDouble();
			this.e7 			= in.readDouble();
			this.e8 			= in.readDouble();
			this.e9 			= in.readDouble();
			this.prevPartition	= in.readLong();	
		}

		@Override
		public void write(DataOutput out) throws IOException 
		{
			out.writeLong(partitionID);
			out.writeLong(windowID);
			out.writeLong(key);
			out.writeDouble(e1);
			out.writeDouble(e2);
			out.writeDouble(e3);
			out.writeDouble(e4);
			out.writeDouble(e5);
			out.writeDouble(e6);
			out.writeDouble(e7);
			out.writeDouble(e8);
			out.writeDouble(e9);
			out.writeLong(prevPartition);
		}

		@Override
		public int compareTo(VectorElem9D o) 
		{

			if (this.e1 != o.getE1())
				return this.e1 < o.getE1() ? -1 : 1;
			else if (this.e2 != o.getE2())
				return this.e2 < o.getE2() ? -1 : 1;
			else if (this.e3 != o.getE3())
				return this.e3 < o.getE3() ? -1 : 1;
			else if (this.e4 != o.getE4())
				return this.e4 < o.getE4() ? -1 : 1;
			else if (this.e5 != o.getE5())
				return this.e5 < o.getE5() ? -1 : 1;
			else if (this.e6 != o.getE6())
				return this.e6 < o.getE6() ? -1 : 1;
			else if (this.e7 != o.getE7())
				return this.e7 < o.getE7() ? -1 : 1;
			else if (this.e8 != o.getE8())
				return this.e8 < o.getE8() ? -1 : 1;
			else if (this.e9 != o.getE9())
				return this.e9 < o.getE9() ? -1 : 1;
			else
				return 0;
		}
		
		//must be used for grouper, abstract class
		public int compareToGrouper(VectorElem9D o)
		{
			if (this.key != o.getKey())
				return this.key < o.getKey() ? -1 : 1;
			else
				return 0;
		}
		
		@Override
		public String toString()
		{
			return (key + " " + e1 + " " + e2 + " " + e3 + " " + e4 + " " + e5 + " " + e6 + " " + e7 + " " + e8 + " " + e9);	
		}
		
		public String toStringPart()
		{
			return (key + " " + e1 + " " + e2 + " " + e3 + " " + e4 + " " + e5 + " " + e6 + " " + e7 + " " + e8 + " " + e9 + " " + partitionID);

		}
		
		public String toStringPrev()
		{
			return (key + " " + e1 + " " + e2 + " " + e3 + " " + e4 + " " + e5 + " " + e6 + " " + e7 + " " + e8 + " " + e9 + " " + prevPartition);
		}
		
	}// end VectorElem9D
	
 	public static class VectorElem16D implements WritableComparable<VectorElem16D>
	{
		private long		partitionID		= -1;
		private long		windowID		= -1;
		private long		key;
		private double 	e1;
		private double 	e2;
		private double 	e3;
		private double 	e4;
		private double 	e5;
		private double 	e6;
		private double 	e7;
		private double 	e8;
		private double 	e9;
		private double 	e10;
		private double 	e11;
		private double 	e12;
		private double 	e13;
		private double 	e14;
		private double 	e15;
		private double 	e16;
		private long		prevPartition 	= -1;
		
		VectorElem16D()
		{
			
		}
		
		VectorElem16D(long key, double e1,double e2, double e3, double e4, double e5, double e6, double e7, double e8, double e9, double e10, 
				double e11, double e12, double e13, double e14, double e15, double e16)
		{
			this.key = key;
			this.e1 = e1;
			this.e2 = e2;
			this.e3 = e3;
			this.e4 = e4;
			this.e5 = e5;
			this.e6 = e6;
			this.e7 = e7;
			this.e8 = e8;
			this.e9 = e9;
			this.e10 = e10;
			this.e11 = e11;
			this.e12 = e12;
			this.e13 = e13;
			this.e14 = e14;
			this.e15 = e15;
			this.e16 = e16;
		}
		
		//this could be an abstract method in an abstract class in, 
		//take a text type and populate the values based off the input
		VectorElem16D(Text text)
		{
			StringTokenizer line = new StringTokenizer(text.toString());
			this.key = Long.parseLong(line.nextToken());
			this.e1 = Double.parseDouble(line.nextToken());
			this.e2 = Double.parseDouble(line.nextToken());
			this.e3 = Double.parseDouble(line.nextToken());
			this.e4 = Double.parseDouble(line.nextToken());
			this.e5 = Double.parseDouble(line.nextToken());
			this.e6 = Double.parseDouble(line.nextToken());
			this.e7 = Double.parseDouble(line.nextToken());
			this.e8 = Double.parseDouble(line.nextToken());
			this.e9 = Double.parseDouble(line.nextToken());
			this.e10 = Double.parseDouble(line.nextToken());
			this.e11 = Double.parseDouble(line.nextToken());
			this.e12 = Double.parseDouble(line.nextToken());
			this.e13 = Double.parseDouble(line.nextToken());
			this.e14 = Double.parseDouble(line.nextToken());
			this.e15 = Double.parseDouble(line.nextToken());
			this.e16 = Double.parseDouble(line.nextToken());
			if(line.hasMoreElements())
				this.prevPartition = Long.parseLong(line.nextToken());
		}
		
		VectorElem16D(String text) 
		{
			StringTokenizer line = new StringTokenizer(text);
			this.key = Long.parseLong(line.nextToken());
			this.e1 = Double.parseDouble(line.nextToken());
			this.e2 = Double.parseDouble(line.nextToken());
			this.e3 = Double.parseDouble(line.nextToken());
			this.e4 = Double.parseDouble(line.nextToken());
			this.e5 = Double.parseDouble(line.nextToken());
			this.e6 = Double.parseDouble(line.nextToken());
			this.e7 = Double.parseDouble(line.nextToken());
			this.e8 = Double.parseDouble(line.nextToken());
			this.e9 = Double.parseDouble(line.nextToken());
			this.e10 = Double.parseDouble(line.nextToken());
			this.e11 = Double.parseDouble(line.nextToken());
			this.e12 = Double.parseDouble(line.nextToken());
			this.e13 = Double.parseDouble(line.nextToken());
			this.e14 = Double.parseDouble(line.nextToken());
			this.e15 = Double.parseDouble(line.nextToken());
			this.e16 = Double.parseDouble(line.nextToken());
			if(line.hasMoreElements())
				this.prevPartition = Long.parseLong(line.nextToken());
		}
		
		public int getSizeInBytes()
		{
			return (( 	Long.SIZE 	+		//partitionID
						Long.SIZE 	+		//windowID
						Long.SIZE 	+		//key
						((Double.SIZE) * 16)	+		//elems
						Long.SIZE 			//prevPartition
					)/8);
			
			
		}
		
		public long getKey()
		{
			return key;
		}
		
		public void setPartitionID(long partitionID) {
			this.partitionID = partitionID;
		}

		public long getPartitionID() {
			return partitionID;
		}

		public long getWindowID() {
			return windowID;
		}

		public void setWindowID(long windowID) {
			this.windowID = windowID;
		}
		
		public long getPrevPartition() {
			return prevPartition;
		}

		public void setPrevPartition(long prevPartition) {
			this.prevPartition = prevPartition;
		}
		
		public double getE1() {
			return e1;
		}

		public void setE1(double e1) {
			this.e1 = e1;
		}

		public double getE2() {
			return e2;
		}

		public void setE2(double e2) {
			this.e2 = e2;
		}

		public double getE3() {
			return e3;
		}

		public void setE3(double e3) {
			this.e3 = e3;
		}

		public double getE4() {
			return e4;
		}

		public void setE4(double e4) {
			this.e4 = e4;
		}

		public double getE5() {
			return e5;
		}

		public void setE5(double e5) {
			this.e5 = e5;
		}

		public double getE6() {
			return e6;
		}

		public void setE6(double e6) {
			this.e6 = e6;
		}

		public double getE7() {
			return e7;
		}

		public void setE7(double e7) {
			this.e7 = e7;
		}

		public double getE8() {
			return e8;
		}

		public void setE8(double e8) {
			this.e8 = e8;
		}

		public double getE9() {
			return e9;
		}

		public void setE9(double e9) {
			this.e9 = e9;
		}

		public double getE10() {
			return e10;
		}

		public void setE10(double e10) {
			this.e10 = e10;
		}

		public double getE11() {
			return e11;
		}

		public void setE11(double e11) {
			this.e11 = e11;
		}

		public double getE12() {
			return e12;
		}

		public void setE12(double e12) {
			this.e12 = e12;
		}

		public double getE13() {
			return e13;
		}

		public void setE13(double e13) {
			this.e13 = e13;
		}

		public double getE14() {
			return e14;
		}

		public void setE14(double e14) {
			this.e14 = e14;
		}

		public double getE15() {
			return e15;
		}

		public void setE15(double e15) {
			this.e15 = e15;
		}

		public double getE16() {
			return e16;
		}

		public void setE16(double e16) {
			this.e16 = e16;
		}


		//would need to be implemented in an abstract class
		public double getDistanceBetween(VectorElem16D o)
		{
			//double deltaX = this.x - o.getX();
			//double deltaY = this.y - o.getY();
			if(size == 16)
			{
				return Math.sqrt(
						Math.pow((this.e1 - o.getE1()), 2) + 
						Math.pow((this.e2 - o.getE2()), 2) + 
						Math.pow((this.e3 - o.getE3()), 2) + 
						Math.pow((this.e4 - o.getE4()), 2) + 
						Math.pow((this.e5 - o.getE5()), 2) + 
						Math.pow((this.e6 - o.getE6()), 2) + 
						Math.pow((this.e7 - o.getE7()), 2) + 
						Math.pow((this.e8 - o.getE8()), 2) + 
						Math.pow((this.e9 - o.getE9()), 2) + 
						Math.pow((this.e10 - o.getE10()), 2) + 
						Math.pow((this.e11 - o.getE11()), 2) + 
						Math.pow((this.e12 - o.getE12()), 2) + 
						Math.pow((this.e13 - o.getE13()), 2) + 
						Math.pow((this.e14 - o.getE14()), 2) + 
						Math.pow((this.e15 - o.getE15()), 2) + 
						Math.pow((this.e16 - o.getE16()), 2)
						) ;
			}
			else if(size == 12)
			{
				return Math.sqrt(
						Math.pow((this.e1 - o.getE1()), 2) + 
						Math.pow((this.e2 - o.getE2()), 2) + 
						Math.pow((this.e3 - o.getE3()), 2) + 
						Math.pow((this.e4 - o.getE4()), 2) + 
						Math.pow((this.e5 - o.getE5()), 2) + 
						Math.pow((this.e6 - o.getE6()), 2) + 
						Math.pow((this.e7 - o.getE7()), 2) + 
						Math.pow((this.e8 - o.getE8()), 2) + 
						Math.pow((this.e9 - o.getE9()), 2) + 
						Math.pow((this.e10 - o.getE10()), 2) + 
						Math.pow((this.e11 - o.getE11()), 2) + 
						Math.pow((this.e12 - o.getE12()), 2)
						) ;
			}
			else if(size == 10)
			{
				return Math.sqrt(
						Math.pow((this.e1 - o.getE1()), 2) + 
						Math.pow((this.e2 - o.getE2()), 2) + 
						Math.pow((this.e3 - o.getE3()), 2) + 
						Math.pow((this.e4 - o.getE4()), 2) + 
						Math.pow((this.e5 - o.getE5()), 2) + 
						Math.pow((this.e6 - o.getE6()), 2) + 
						Math.pow((this.e7 - o.getE7()), 2) + 
						Math.pow((this.e8 - o.getE8()), 2) + 
						Math.pow((this.e9 - o.getE9()), 2) + 
						Math.pow((this.e10 - o.getE10()), 2)
						) ;
			}
			else if(size == 9)
			{
				return Math.sqrt(
						Math.pow((this.e1 - o.getE1()), 2) + 
						Math.pow((this.e2 - o.getE2()), 2) + 
						Math.pow((this.e3 - o.getE3()), 2) + 
						Math.pow((this.e4 - o.getE4()), 2) + 
						Math.pow((this.e5 - o.getE5()), 2) + 
						Math.pow((this.e6 - o.getE6()), 2) + 
						Math.pow((this.e7 - o.getE7()), 2) + 
						Math.pow((this.e8 - o.getE8()), 2) + 
						Math.pow((this.e9 - o.getE9()), 2) 
						) ;
			}
			else if(size == 8)
			{
				return Math.sqrt(
						Math.pow((this.e1 - o.getE1()), 2) + 
						Math.pow((this.e2 - o.getE2()), 2) + 
						Math.pow((this.e3 - o.getE3()), 2) + 
						Math.pow((this.e4 - o.getE4()), 2) + 
						Math.pow((this.e5 - o.getE5()), 2) + 
						Math.pow((this.e6 - o.getE6()), 2) + 
						Math.pow((this.e7 - o.getE7()), 2) + 
						Math.pow((this.e8 - o.getE8()), 2) 
						) ;
			}
			else if(size == 6)
			{
				return Math.sqrt(
						Math.pow((this.e1 - o.getE1()), 2) + 
						Math.pow((this.e2 - o.getE2()), 2) + 
						Math.pow((this.e3 - o.getE3()), 2) + 
						Math.pow((this.e4 - o.getE4()), 2) + 
						Math.pow((this.e5 - o.getE5()), 2) + 
						Math.pow((this.e6 - o.getE6()), 2)
						) ;
			}
			else if(size == 4)
			{
				return Math.sqrt(
						Math.pow((this.e1 - o.getE1()), 2) + 
						Math.pow((this.e2 - o.getE2()), 2) + 
						Math.pow((this.e3 - o.getE3()), 2) + 
						Math.pow((this.e4 - o.getE4()), 2)
						) ;
			}
			else
				return Integer.MAX_VALUE;
		}

		@Override
		public void readFields(DataInput in) throws IOException 
		{
			this.partitionID 	= in.readLong();
			this.windowID 		= in.readLong();
			this.key 			= in.readLong();
			this.e1 			= in.readDouble();
			this.e2 			= in.readDouble();
			this.e3 			= in.readDouble();
			this.e4 			= in.readDouble();
			this.e5 			= in.readDouble();
			this.e6 			= in.readDouble();
			this.e7 			= in.readDouble();
			this.e8 			= in.readDouble();
			this.e9 			= in.readDouble();
			this.e10 			= in.readDouble();
			this.e11 			= in.readDouble();
			this.e12 			= in.readDouble();
			this.e13 			= in.readDouble();
			this.e14 			= in.readDouble();
			this.e15 			= in.readDouble();
			this.e16 			= in.readDouble();
			this.prevPartition	= in.readLong();	
		}

		@Override
		public void write(DataOutput out) throws IOException 
		{
			out.writeLong(partitionID);
			out.writeLong(windowID);
			out.writeLong(key);
			out.writeDouble(e1);
			out.writeDouble(e2);
			out.writeDouble(e3);
			out.writeDouble(e4);
			out.writeDouble(e5);
			out.writeDouble(e6);
			out.writeDouble(e7);
			out.writeDouble(e8);
			out.writeDouble(e9);
			out.writeDouble(e10);
			out.writeDouble(e11);
			out.writeDouble(e12);
			out.writeDouble(e13);
			out.writeDouble(e14);
			out.writeDouble(e15);
			out.writeDouble(e16);
			out.writeLong(prevPartition);
		}

		@Override
		public int compareTo(VectorElem16D o) 
		{

			if (this.e1 != o.getE1())
				return this.e1 < o.getE1() ? -1 : 1;
			else if (this.e2 != o.getE2())
				return this.e2 < o.getE2() ? -1 : 1;
			else if (this.e3 != o.getE3())
				return this.e3 < o.getE3() ? -1 : 1;
			else if (this.e4 != o.getE4())
				return this.e4 < o.getE4() ? -1 : 1;
			else if (this.e5 != o.getE5())
				return this.e5 < o.getE5() ? -1 : 1;
			else if (this.e6 != o.getE6())
				return this.e6 < o.getE6() ? -1 : 1;
			else if (this.e7 != o.getE7())
				return this.e7 < o.getE7() ? -1 : 1;
			else if (this.e8 != o.getE8())
				return this.e8 < o.getE8() ? -1 : 1;
			else if (this.e9 != o.getE9())
				return this.e9 < o.getE9() ? -1 : 1;
			else if (this.e10 != o.getE10())
				return this.e10 < o.getE10() ? -1 : 1;
			else if (this.e11 != o.getE11())
				return this.e11 < o.getE11() ? -1 : 1;
			else if (this.e12 != o.getE12())
				return this.e12 < o.getE12() ? -1 : 1;
			else if (this.e13 != o.getE13())
				return this.e13 < o.getE13() ? -1 : 1;
			else if (this.e14 != o.getE14())
				return this.e14 < o.getE14() ? -1 : 1;
			else if (this.e15 != o.getE15())
				return this.e15 < o.getE15() ? -1 : 1;
			else if (this.e16 != o.getE16())
				return this.e16 < o.getE16() ? -1 : 1;
			else
				return 0;
		}
		
		//must be used for grouper, abstract class
		public int compareToGrouper(VectorElem16D o)
		{
			if (this.key != o.getKey())
				return this.key < o.getKey() ? -1 : 1;
			else
				return 0;
		}
		
		@Override
		public String toString()
		{
			return (key + " " + e1 + " " + e2 + " " + e3 + " " + e4 + " " + e5 + " " + e6 + " " + e7 + " " + e8 + " " + e9 + " " + e10 + " " + e11 + " " + e12 + " " + e13 + " " + e14 + " " + e15 + " " + e16);	
		}
		
		public String toStringPart()
		{
			return (key + " " + e1 + " " + e2 + " " + e3 + " " + e4 + " " + e5 + " " + e6 + " " + e7 + " " + e8 + " " + e9 + " " + e10 + " " + e11 + " " + e12 + " " + e13 + " " + e14 + " " + e15 + " " + e16 + " " + partitionID);

		}
		
		public String toStringPrev()
		{
			return (key + " " + e1 + " " + e2 + " " + e3 + " " + e4 + " " + e5 + " " + e6 + " " + e7 + " " + e8 + " " + e9 + " " + e10 + " " + e11 + " " + e12 + " " + e13 + " " + e14 + " " + e15 + " " + e16 + " " + prevPartition);
		}
		
	}// end VectorElem16D

 	public static class StringElem implements WritableComparable<StringElem>
	{
		private long		partitionID		= -1;
		private long		windowID		= -1;
		private long key;
		private String authors;
		private String editor;
		private String title;
		private String booktitle;
		private String pages;
		private String year;
		private String address;
		private String journal;
		private String volume;
		private String number;
		private String month;
		private String url;
		private String ee;
		private String cdrom;
		private String cite;
		private String publisher;
		private String note;
		private String crossref;
		private String isbn;
		private String series;
		private String school;
		private String chapter;
		private String recordType;
		private long		prevPartition 	= -1;
		
		StringElem()
		{
			
		}
		
		//this could be an abstract method in an abstract class in, 
		//take a text type and populate the values based off the input
		StringElem(Text text)
		{
            StringTokenizer line = new StringTokenizer(text.toString(), "|");
			
			key = Long.parseLong(line.nextToken());
            authors = line.nextToken();
            editor = line.nextToken();
            title = line.nextToken();
            booktitle = line.nextToken();
            pages = line.nextToken();
            year = line.nextToken();
            address = line.nextToken();
            journal = line.nextToken();
            volume = line.nextToken();
            number = line.nextToken();
            month = line.nextToken();
            url = line.nextToken();
            ee = line.nextToken();
            cdrom = line.nextToken();
            cite = line.nextToken();
            publisher = line.nextToken();
            note = line.nextToken();
            crossref = line.nextToken();
            isbn = line.nextToken();
            series = line.nextToken();
            school = line.nextToken();
            chapter = line.nextToken();
            recordType = line.nextToken();
            if(line.hasMoreTokens())
            	prevPartition = Long.parseLong(line.nextToken());
		}
		
		StringElem(String text) 
		{
			StringTokenizer line = new StringTokenizer(text, "|");
			key = Long.parseLong(line.nextToken());
            authors = line.nextToken();
            editor = line.nextToken();
            title = line.nextToken();
            booktitle = line.nextToken();
            pages = line.nextToken();
            year = line.nextToken();
            address = line.nextToken();
            journal = line.nextToken();
            volume = line.nextToken();
            number = line.nextToken();
            month = line.nextToken();
            url = line.nextToken();
            ee = line.nextToken();
            cdrom = line.nextToken();
            cite = line.nextToken();
            publisher = line.nextToken();
            note = line.nextToken();
            crossref = line.nextToken();
            isbn = line.nextToken();
            series = line.nextToken();
            school = line.nextToken();
            chapter = line.nextToken();
            recordType = line.nextToken();
            if(line.hasMoreTokens())
            	prevPartition = Long.parseLong(line.nextToken());
		}
		
		public int getSizeInBytes()
		{
			return ( 	325 +	//size of string
						(Long.SIZE 	+		//partitionID
						Long.SIZE 	+		//windowID
						Long.SIZE 	+		//key
						Long.SIZE 			//prevPartition
						)/8
					);
			
			
		}
		
		public long getKey()
		{
			return key;
		}
		
		public void setPartitionID(long partitionID) {
			this.partitionID = partitionID;
		}

		public long getPartitionID() {
			return partitionID;
		}

		public long getWindowID() {
			return windowID;
		}

		public void setWindowID(long windowID) {
			this.windowID = windowID;
		}
		
		public long getPrevPartition() {
			return prevPartition;
		}

		public void setPrevPartition(long prevPartition) {
			this.prevPartition = prevPartition;
		}
		
		public String getValue() {
			return title;
		}

		public void setValue(String value) {
			this.title = value;
		}

		//would need to be implemented in an abstract class
		public double getDistanceBetween(StringElem o)
		{
			char[] c1 = this.title.toCharArray();
            char[] c2 = o.getValue().toCharArray();

            int[][] d = new int[c1.length + 1][c2.length + 1];

            for(int i = 0; i <= c1.length; i++)
                    d[i][0] = i;
            for(int j = 0; j <= c2.length; j++)
                    d[0][j] = j;

            for(int i = 1; i <= c1.length; i++)
            {
                    for(int j = 1; j <= c2.length; j++)
                    {
                                    d[i][j] = Math.min(
                                                            (d[i-1][j] + 1), Math.min(		//deletion
                                                            (d[i][j-1] + 1), 			//insertion
                                                            (d[i-1][j-1]) + ((c1[i - 1] == c2[j - 1]) ? 0 : 1)) //substitution
                                                      );
                    }
            }
            return d[c1.length][c2.length];
		}

		@Override
		public void readFields(DataInput in) throws IOException 
		{
			partitionID = in.readLong();
			windowID = in.readLong();
			key = in.readLong();
			authors = in.readLine();
			editor = in.readLine();
			title = in.readLine();
			booktitle = in.readLine();
			pages = in.readLine();
			year = in.readLine();
			address = in.readLine();
			journal = in.readLine();
			volume = in.readLine();
			number = in.readLine();
			month = in.readLine();
			url = in.readLine();
			ee = in.readLine();
			cdrom = in.readLine();
			cite = in.readLine();
			publisher = in.readLine();
			note = in.readLine();
			crossref = in.readLine();
			isbn = in.readLine();
			series = in.readLine();
			school = in.readLine();
			chapter = in.readLine();
			recordType = in.readLine();
			prevPartition = in.readLong();
		}
		
		@Override
		public void write(DataOutput out) throws IOException 
		{
			out.writeLong(partitionID);
			out.writeLong(windowID);
			out.writeLong(key);
			out.writeBytes(authors + '\n');
			out.writeBytes(editor + '\n');
			out.writeBytes(title + '\n');
			out.writeBytes(booktitle + '\n');
			out.writeBytes(pages + '\n');
			out.writeBytes(year + '\n');
			out.writeBytes(address + '\n');
			out.writeBytes(journal + '\n');
			out.writeBytes(volume + '\n');
			out.writeBytes(number + '\n');
			out.writeBytes(month + '\n');
			out.writeBytes(url + '\n');
			out.writeBytes(ee + '\n');
			out.writeBytes(cdrom + '\n');
			out.writeBytes(cite + '\n');
			out.writeBytes(publisher + '\n');
			out.writeBytes(note + '\n');
			out.writeBytes(crossref + '\n');
			out.writeBytes(isbn + '\n');
			out.writeBytes(series + '\n');
			out.writeBytes(school + '\n');
			out.writeBytes(chapter + '\n');
			out.writeBytes(recordType + '\n');
			out.writeLong(prevPartition);
		}

		@Override
		public int compareTo(StringElem o) 
		{
			return title.compareTo(o.getValue());
		}
		
		//must be used for grouper, abstract class
		public int compareToGrouper(StringElem o)
		{
			if (this.key != o.getKey())
				return this.key < o.getKey() ? -1 : 1;
			else
				return 0;
		}
		
		@Override
		public String toString()
		{ 
			char seperator = '|';
			return (key + "" + seperator +
					authors + seperator +
					editor + seperator +
					title + seperator +
					booktitle + seperator +
					pages + seperator +
					year + seperator +
					address + seperator +
					journal + seperator +
					volume + seperator +
					number + seperator +
					month + seperator +
					url + seperator +
					ee + seperator +
					cdrom + seperator + 
					cite + seperator +
					publisher + seperator +
					note + seperator +
					crossref + seperator +
					isbn + seperator +
					series + seperator +
					school + seperator +
					chapter + seperator +
					recordType);	
		}
		
		public String toStringPart()
		{
			char seperator = '|';
			return (key + "" + seperator +
					authors + seperator +
					editor + seperator +
					title + seperator +
					booktitle + seperator +
					pages + seperator +
					year + seperator +
					address + seperator +
					journal + seperator +
					volume + seperator +
					number + seperator +
					month + seperator +
					url + seperator +
					ee + seperator +
					cdrom + seperator + 
					cite + seperator +
					publisher + seperator +
					note + seperator +
					crossref + seperator +
					isbn + seperator +
					series + seperator +
					school + seperator +
					chapter + seperator +
					recordType + seperator + partitionID);
		}
		
		public String toStringPrev()
		{
			char seperator = '|';
			return (key + "" + seperator +
					authors + seperator +
					editor + seperator +
					title + seperator +
					booktitle + seperator +
					pages + seperator +
					year + seperator +
					address + seperator +
					journal + seperator +
					volume + seperator +
					number + seperator +
					month + seperator +
					url + seperator +
					ee + seperator +
					cdrom + seperator + 
					cite + seperator +
					publisher + seperator +
					note + seperator +
					crossref + seperator +
					isbn + seperator +
					series + seperator +
					school + seperator +
					chapter + seperator +
					recordType + seperator + prevPartition);
		}
		
	}// end StringElem
 	
 	public static class CloudJoinKey implements WritableComparable<CloudJoinKey>
 	{
 		
 		long partitionID		= -1;
 		long windowID			= -1;
 		long prevItrPartion	 	= -1;
 		
 		CloudJoinKey()
 		{
 			System.out.println("Why am I seeing this?");
 		}
 		
 		CloudJoinKey(long partitionID)
 		{
 			this.partitionID = partitionID;
 		}
 		
 		CloudJoinKey(long partitionID, long windowID)
 		{
 			this.partitionID = partitionID;
 			this.windowID = windowID;
 		}

 		CloudJoinKey(long partitionID, long windowID, long prevItrPartion)
 		{
 			this.partitionID = partitionID;
 			this.windowID = windowID;
 			this.prevItrPartion = prevItrPartion;
 		}		
 		
		public long getPartitionID() {
			return partitionID;
		}

		public void setPartitionID(long partitionID) {
			this.partitionID = partitionID;
		}

		public long getWindowID() {
			return windowID;
		}

		public void setWindowID(long windowID) {
			this.windowID = windowID;
		}

		public long getPrevItrPartion() {
			return prevItrPartion;
		}

		public void setPrevItrPartion(long prevItrPartion) {
			this.prevItrPartion = prevItrPartion;
		}

		@Override
		public void readFields(DataInput in) throws IOException {
			partitionID = in.readLong();
			windowID = in.readLong();
			prevItrPartion = in.readLong();
		}

		@Override
		public void write(DataOutput out) throws IOException {
			out.writeLong(partitionID);
			out.writeLong(windowID);
			out.writeLong(prevItrPartion);
		}

		@Override
		public int compareTo(CloudJoinKey o) 
		{
			if(this.partitionID != o.getPartitionID())
				return this.partitionID < o.getPartitionID() ? -1 : 1;
			else if(this.windowID != o.getWindowID())
				return this.windowID < o.getWindowID() ? -1 : 1;
			else if(this.prevItrPartion != o.getPrevItrPartion())
				return this.prevItrPartion < o.getPrevItrPartion() ? -1 : 1;
			else
				return 0;
		}
		
		public String toString()
		{
			return "PID: " + partitionID + " WID: " + windowID + " PIP: " + prevItrPartion;	
		}
 	}
 	
	public static class CloudJoinBaseMapper extends Mapper<LongWritable, Text, CloudJoinKey, StringElem>
	{
		ArrayList<StringElem> pivPts;
		
		@Override
		protected void setup(Context context) {
			log.setLevel(Level.ALL);
			log.debug("BM we have started configure?");
		    try {
		      
		    	Path[] cacheFiles = DistributedCache.getLocalCacheFiles(context.getConfiguration());
		    	
		    	log.debug("BM cacheFiles length: " + cacheFiles.length);
		      if (null != cacheFiles && cacheFiles.length == 1) 
		      {
		    	  log.debug("BM Did we get in here?");
		    	  	Path cachePath = cacheFiles[0];
		    	  	log.debug("BM cachePath: " + cachePath.toString());
		            loadPivPts(cachePath);
		      }
		      else
		      {
		    	  log.debug("BM Error accessing distributed cache");
		      }
		    } catch (IOException ioe) {
		    	log.debug("BM IOException reading from distributed cache");
		    	log.debug(ioe.toString());
		    }
		  }

		void loadPivPts(Path cachePath) throws IOException {
		    // note use of regular java.io methods here - this is a local file now
		    BufferedReader lineReader = new BufferedReader(new FileReader(cachePath.toString()));
		    try {
		      String line;
		      this.pivPts = new ArrayList<StringElem>();
		      while ((line = lineReader.readLine()) != null) 
		      {
		    	  //parse the line into a StringElem object
		    	  this.pivPts.add(new StringElem(line));
		      }
		      
		    } finally {
		      lineReader.close();
		    }
		  }
	
		public void map(LongWritable inputKey, Text inputValue, Context context) throws IOException, InterruptedException
		{
			log.debug("===================================Base Map Begin===================================");
			String s;
			s = context.getConfiguration().get("eps", "");
			if("".equals(s))
			{
				log.error("BM Error: Unable to get eps from configuration");
			}
			double eps = Double.parseDouble(s);
			
			StringElem elem;
			try
			{
				elem = new StringElem(inputValue);				//creates element that is being worked on in this map
			}
			catch(Exception e)
			{
				log.debug("we gots an error...");
				log.debug(inputValue.toString());
				e.printStackTrace();
				return;
			}

			StringElem  pivot_curr;										//tracks the current pivot point
			StringElem  pivot_best;										//current best pivot point/partition
			double min_distance;											//current min distance between elem and pivot_best
			
			Iterator<StringElem> pivItr = pivPts.iterator();
			log.debug("BM How long is pivPts: " + pivPts.size());
			if(pivItr.hasNext())
			{
				pivot_best = pivItr.next();									//establish a first pivot point
				min_distance = elem.getDistanceBetween(pivot_best);
			}
			else
			{
				log.error("BM Error: Distributed cache is empty");	//error if none found
				return;
			}
			while(pivItr.hasNext())											//check all partitions point and find best match
			{
				pivot_curr = pivItr.next();
				double curr_distance = elem.getDistanceBetween(pivot_curr);
				if(curr_distance < min_distance )							//if the current partition is closer than the current best
				{
					pivot_best = pivot_curr;								//update pivot_best
					min_distance = curr_distance;							//update current distance
				}
			}
			elem.setPartitionID(pivot_best.getKey());						//set element native partition (may not be needed)
			CloudJoinKey cjk = new CloudJoinKey(pivot_best.getKey());		//creates a key for writing to the context
			context.write(cjk, elem);										//writes to context the key value pair

			String dbg = "BM----------------------------------------------------------------------" + '\n';
			dbg += "BM Elem: " + elem.toStringPart()  + '\n';
			dbg += "BM CJK: " + cjk.toString()  + '\n';
			dbg += "BM----------------------------------------------------------------------" + '\n';
			log.debug(dbg);


			pivItr = pivPts.iterator();										//check if the elem is in a window of any of the other partitions
			while(pivItr.hasNext())
			{
				pivot_curr = pivItr.next();									//get a current working pivot
				if(pivot_curr.getKey()==pivot_best.getKey())
				{
					continue;												//if the current pivot is the same as the best match, 
																			//continue onto next pivot
				}
				double curr_distance = elem.getDistanceBetween(pivot_curr);//get the distance between the current pivot and the elem
				
				if(((curr_distance-min_distance)/2) <= eps)					//check if it's in in the range of the window
				{
					//add to window partition
					cjk.setWindowID(pivot_curr.getKey());					//set the window id on the keyto the key key of the current pivot
					context.write(cjk, elem);
					
					dbg = "BM window----------------------------------------------------------------------" + '\n';
					dbg += "BM window Elem: " + elem.toStringPart()  + '\n';
					dbg += "BM window CJK: " + cjk.toString()  + '\n';
					dbg += "BM window----------------------------------------------------------------------" + '\n';
					log.debug(dbg);
				}
			}
			log.debug("===================================Base Map End===================================");
		}//end map
		
	}//end CloudJoinBaseMapper
		
	public static class CloudJoinWindowMapper extends Mapper<LongWritable, Text, CloudJoinKey, StringElem>
	{
		ArrayList<StringElem> pivPts;
		
		@Override
		protected void setup(Context context) {
			log.setLevel(Level.ALL);
			log.debug("WM we have started configure?");
		    try {
		      Path[] cacheFiles = DistributedCache.getLocalCacheFiles(context.getConfiguration());
		    	
		    	log.debug("WM cacheFiles length: " + cacheFiles.length);
		      if (null != cacheFiles && cacheFiles.length == 1) 
		      {
		    	  log.debug("WM Did we get in here?");
		    	  	Path cachePath = cacheFiles[0];
		    	  	log.debug("WM cachePath: " + cachePath.toString());
		            loadPivPts(cachePath);
		      }
		      else
		      {
		    	  log.debug("WM Error accessing distributed cache");
		      }
		    } catch (IOException ioe) {
		    	log.debug("WM IOException reading from distributed cache");
		    	log.debug(ioe.toString());
		    }
		  }

		void loadPivPts(Path cachePath) throws IOException {
		    // note use of regular java.io methods here - this is a local file now
		    BufferedReader lineReader = new BufferedReader(new FileReader(cachePath.toString()));
		    try {
		      String line;
		      this.pivPts = new ArrayList<StringElem>();
		      while ((line = lineReader.readLine()) != null) 
		      {
		    	  //parse the line into a StringElem object
		    	  this.pivPts.add(new StringElem(line));
		      }
		      
		    } finally {
		      lineReader.close();
		    }
		  }
			
		public void map(LongWritable inputKey, Text inputValue, Context context) throws IOException, InterruptedException
		{
			log.debug("===================================Window Map Begin===================================");
			String s;
			s = context.getConfiguration().get("eps", "");
			if("".equals(s))
			{
				System.out.println("Error: Unable to get eps from configuration");
			}
			double eps = Double.parseDouble(s);
			s = context.getConfiguration().get("W", "");
			if("".equals(s))
			{
				System.out.println("Error: Unable to get memory from configuration");
			}
			long w = Long.parseLong(s);
			s = context.getConfiguration().get("V", "");
			if("".equals(s))
			{
				System.out.println("Error: Unable to get itr from configuration");
			}
			long v = Long.parseLong(s);
						
			StringElem elem;
			try
			{
				elem = new StringElem(inputValue);				//creates element that is being worked on in this map
			}
			catch(Exception e)
			{
				log.debug(inputValue.toString());
				return;
			}
			
			StringElem  pivot_curr;										//tracks the current pivot point
			StringElem  pivot_best;										//current best pivot point/partition
			double min_distance;											//current min distance between elem and pivot_best
			
			Iterator<StringElem> pivItr = pivPts.iterator();
			if(pivItr.hasNext())
			{
				pivot_best = pivItr.next();									//establish a first pivot point
				min_distance = elem.getDistanceBetween(pivot_best);
			}
			else
			{
				System.out.println("Error: Distributed cache is empty");	//error if none found
				return;
			}
			while(pivItr.hasNext())											//check all partitions point and find best match
			{
				pivot_curr = pivItr.next();
				double curr_distance = elem.getDistanceBetween(pivot_curr);
				if(curr_distance < min_distance )							//if the current partition is closer than the current best
				{
					pivot_best = pivot_curr;								//update pivot_best
					min_distance = curr_distance;							//update current distance
				}
			}
			
																			//creates a key for writing to the context
			CloudJoinKey cjk = new CloudJoinKey(pivot_best.getKey(), -1, elem.getPrevPartition());
			elem.setPartitionID(pivot_best.getKey());						//set element native partition (may not be needed)
			context.write(cjk, elem);										//writes to context the key value pair
			
			String dbg = "WM----------------------------------------------------------------------" + '\n';
			dbg += "WM Elem: " + elem.toStringPart()  + '\n';
			dbg += "WM CJK: " + cjk.toString()  + '\n';
			dbg += "WM----------------------------------------------------------------------" + '\n';
			log.debug(dbg);
	
			pivItr = pivPts.iterator();										//check if the elem is in a window of any of the other partitions
			while(pivItr.hasNext())
			{
				pivot_curr = pivItr.next();									//get a current working pivot
				if(pivot_curr.getKey()==pivot_best.getKey())
				{
					continue;												//if the current pivot is the same as the best match, 
																			//continue onto next pivot
				}
				double curr_distance = elem.getDistanceBetween(pivot_curr);//get the distance between the current pivot and the elem
				
				if(((curr_distance-min_distance)/2) <= eps)					//check if it's in in the range of the window
				{
					//add to window partition
					cjk.setWindowID(pivot_curr.getKey());					//set the window id on the keyto the key key of the current pivot
					context.write(cjk, elem);
					
					dbg = "WM window----------------------------------------------------------------------" + '\n';
					dbg += "WM window Elem: " + elem.toStringPart()  + '\n';
					dbg += "WM window CJK: " + cjk.toString()  + '\n';
					dbg += "WM window----------------------------------------------------------------------" + '\n';
					log.debug(dbg);
				}
			}
			log.debug("===================================Window Map End===================================");
		}//end map
		
	}//end CloudJoinWindowMapper

	public static class CloudJoinBasePartitioner extends Partitioner<CloudJoinKey, StringElem>
	{
		@Override
		public int getPartition(CloudJoinKey key, StringElem value, int numP) {
			log.setLevel(Level.ALL);
			
			if(  key.getWindowID() != (-1)  )
			{//if it's a window partition, send to window partition
				long min = Math.min(key.getPartitionID(), key.getWindowID());
				long max = Math.max(key.getPartitionID(), key.getWindowID());
				log.debug("PP Key PID: " + key.getPartitionID() + " Partition: " + (int)((min*157 + max*127) % numP));
				return (int)((min*157 + max*127) % numP);
			}
			else
			{//if it's it's native partition, send to it's native partition
				
				log.debug("PW Key PID: " + key.getPartitionID() + " Partition: " + (int)((key.getPartitionID()*149) % numP));
				return (int)((key.getPartitionID()*149) % numP);
			}
		}
		
	}
		
	public static class CloudJoinWindowPartitioner extends Partitioner<CloudJoinKey, StringElem> implements Configurable
	{
		private Configuration conf;
		
		public int getPartition(CloudJoinKey key, StringElem value, int numP) 
		{
			
			//we should perform a check here and make sure we get values back
			long W = Long.parseLong(conf.get("W"));
			long V = Long.parseLong(conf.get("V"));
			
			if(key.getWindowID() == -1)	
			{
				return (int)((key.getPartitionID() * 163) % numP);
			}
			else
			{
				long min = Math.min(key.getPartitionID(), key.getWindowID());
				long max = Math.max(key.getPartitionID(), key.getWindowID());
				
				if(  ( (key.getPartitionID()<key.getWindowID()) && (key.getPrevItrPartion()==W) ) ||
					 ( (key.getPartitionID()>key.getWindowID()) && (key.getPrevItrPartion()==V) ) )
				{
					return (int)((min*179 + max*197) % numP);
				}
				else
				{
					return (int)((min*173 + max*181) % numP);
				}
			}
		}
		
		@Override
		public Configuration getConf() {
			return conf;
		}

		@Override
		public void setConf(Configuration conf) {
			this.conf = conf;
		}
		
	}
	
	public static class CloudJoinBaseGrouper implements RawComparator<CloudJoinKey>
	{
		
		@Override
		public int compare(CloudJoinKey key1, CloudJoinKey key2)
	    {
			
			long 	part1	= key1.getPartitionID();
			long	win1	= key1.getWindowID();
			long	part2	= key2.getPartitionID();
			long	win2 	= key2.getWindowID();
			if (win1==-1 && win2==-1)
			{//p-p
				if(part1==part2) return 0;
				else return (part1 < part2 ? -1 : 1);
			}
			else if (win1==-1)
			{ //p-w
				return -1;
			}
			else if (win2==-1)
			{ //w-p
				return 1;	
			}
			else
			{//w-w
				long minOfo1 = Math.min(part1, win1);
				long maxOfo1 = Math.max(part1, win1);
				
				long minOfo2 = Math.min(part2, win2);
				long maxOfo2 = Math.max(part2, win2);
				
				if (minOfo1==minOfo2)
				{
					if(maxOfo1==maxOfo2)
						return 0;
					else
						return (maxOfo1 < maxOfo2 ? -1:1);
				}
				else
					return (minOfo1 < minOfo2 ? -1:1);
			}
	 }

		@Override
	    public int compare(	byte[] b1, int s1, int l1, 
	    						byte[] b2, int s2, int l2) 
		{
			
			long part1 = WritableComparator.readLong(b1, s1);
			long win1 = WritableComparator.readLong(b1, s1 + (Long.SIZE / 8));
			long part2 = WritableComparator.readLong(b2, s2);
			long win2 = WritableComparator.readLong(b2, s2 + (Long.SIZE / 8));
			
			
			if (win1==-1 && win2==-1)
			{//p-p
				if(part1==part2) return 0;
				else return (part1 < part2 ? -1 : 1);
			}
			else if (win1==-1)
			{ //p-w
				return -1;
			}
			else if (win2==-1)
			{ //w-p
				return 1;	
			}
			else
			{//w-w
				long minOfo1 = Math.min(part1, win1); //System.out.println("CJBG_BS: minOfo1:" + minOfo1);
				long maxOfo1 = Math.max(part1, win1); //System.out.println("CJBG_BS: maxOfo1:" + maxOfo1);
				
				long minOfo2 = Math.min(part2, win2); //System.out.println("CJBG_BS: minOfo2:" + minOfo2);
				long maxOfo2 = Math.max(part2, win2); //System.out.println("CJBG_BS: maxOfo2:" + maxOfo2);
				
				if (minOfo1==minOfo2)
				{
					if(maxOfo1==maxOfo2)
						return 0;
					else
						return (maxOfo1 < maxOfo2 ? -1:1);
				}
				else
					return (minOfo1 < minOfo2 ? -1:1);
			}
	    }
		
	}
	
	public static class CloudJoinWindowGrouper implements RawComparator<CloudJoinKey>, Configurable
	{
		private Configuration conf;
		
		@Override
		public int compare(CloudJoinKey key1, CloudJoinKey key2)
	    {
			long 	part1		= key1.getPartitionID();
			long	win1		= key1.getWindowID();
			long	part2		= key2.getPartitionID();
			long	win2 		= key2.getWindowID();
			
			if( (win1==-1) && (win2==-1) )//p-p
			{
				if(part1==part2)
				{
					return 0;
				}
				else
				{
					return (part1<part2? -1:1);
				}
			}
			else if( (win1==-1) && (win2!=-1) ) //p-w
			{
				return -1;
			}
			else if ((win1!=-1) && (win2==-1)) //w-p
			{
				return 1;
			}
			else
			{
				
				
				long min1 = Math.min(part1, win1);
				long max1 = Math.max(part1, win1);
				long min2 = Math.min(part2, win2);
				long max2 = Math.max(part2, win2);
				
				if ( !((min1==min2) && (max1==max2)) )
				{
					if(min1==min2)
						return (max1<max2 ? -1:1);
					else
						return (min1<min2 ? -1:1);
				}
				else
				{
					long	prevPart1	= key1.getPrevItrPartion();
					long	prevPart2	= key2.getPrevItrPartion();
					//we should perform a check here and make sure we get values back
					long W = Long.parseLong(conf.get("W"));
					long V = Long.parseLong(conf.get("V"));
					
					if ( (part1==part2) && (prevPart1==prevPart2) )
						return 0;
					else if (part1==part2)
					{
						if (part1<win1)
						{
							if( (prevPart1==V) && (prevPart2==W) )
								return -1;
							else 
								return 1;
						}
						else
						{
							if( (prevPart1==V) && (prevPart2==W) )
								return 1;
							else
								return -1;
						}
					}
					else if (prevPart1==prevPart2)
					{
						if (prevPart1==V)
						{
							if (part1<win1)
								return -1;
							else
								return 1;
						}
						else
						{
							if (part1<win1)
								return 1;
							else
								return -1;
						}
					}
					else
						return 0;
				}
			}
			
	 }// end compare


	    public int compare(	byte[] b1, int s1, int l1, 
	    						byte[] b2, int s2, int l2) 
		{
			
			//this is going to be needing some optimization but this is the easiest way to code this right now.
			
			long part1 = WritableComparator.readLong(b1, s1);
			long win1 = WritableComparator.readLong(b1, s1 + (Long.SIZE / 8));
			long part2 = WritableComparator.readLong(b2, s2);
			long win2 = WritableComparator.readLong(b2, s2 + (Long.SIZE / 8));
		
			if( (win1==-1) && (win2==-1) )//p-p
			{
				if(part1==part2)
				{
					return 0;
				}
				else
				{
					return (part1<part2? -1:1);
				}
			}
			else if( (win1==-1) && (win2!=-1) ) //p-w
			{
				return -1;
			}
			else if ((win1!=-1) && (win2==-1)) //w-p
			{
				return 1;
			}
			else
			{
				
				
				long min1 = Math.min(part1, win1);
				long max1 = Math.max(part1, win1);
				long min2 = Math.min(part2, win2);
				long max2 = Math.max(part2, win2);
				
				if ( !((min1==min2) && (max1==max2)) )
				{
					if(min1==min2)
						return (max1<max2 ? -1:1);
					else
						return (min1<min2 ? -1:1);
				}
				else
				{
					long	prevPart1	= WritableComparator.readLong(b1, s1 + (Long.SIZE + Long.SIZE) / 8);
					long	prevPart2	= WritableComparator.readLong(b2, s2 + (Long.SIZE + Long.SIZE) / 8);
					
					//we should perform a check here and make sure we get values back
					long W = Long.parseLong(conf.get("W"));
					long V = Long.parseLong(conf.get("V"));
					
					if ( (part1==part2) && (prevPart1==prevPart2) )
						return 0;
					else if (part1==part2)
					{
						if (part1<win1)
						{
							if( (prevPart1==V) && (prevPart2==W) )
								return -1;
							else 
								return 1;
						}
						else
						{
							if( (prevPart1==V) && (prevPart2==W) )
								return 1;
							else
								return -1;
						}
					}
					else if (prevPart1==prevPart2)
					{
						if (prevPart1==V)
						{
							if (part1<win1)
								return -1;
							else
								return 1;
						}
						else
						{
							if (part1<win1)
								return 1;
							else
								return -1;
						}
					}
					else
						return 0;
				}
			}
		}
	    
	    
	    
		@Override
		public Configuration getConf() {
			return conf;
		}

		@Override
		public void setConf(Configuration conf) {
			this.conf = conf;
		}
		
	}
	
	public static class CloudJoinBaseReducer extends Reducer<CloudJoinKey, StringElem, Text, Text>
	{

		public void reduce(CloudJoinKey key, Iterable<StringElem> values, Context context)
		{
			//log.setLevel(Level.OFF);
			log.debug("===================================Base Reduce Begin===================================");
			String s;
			s = context.getConfiguration().get("eps", "");
			if("".equals(s))
			{
				log.error("BR Error: Unable to get eps from configuration");
			}
			double eps = Double.parseDouble(s);
			s = context.getConfiguration().get("memory", "");
			if("".equals(s))
			{
				log.error("BR Error: Unable to get memory from configuration");
			}
			long memory = Long.parseLong(s);
			s = context.getConfiguration().get("itr", "");
			if("".equals(s))
			{
				log.error("BR Error: Unable to get itr from configuration");
			}
			long itr = Long.parseLong(s);
			String outDir = context.getConfiguration().get("outDir", "");
			if("".equals(outDir))
			{
				log.error("BR Error: Unable to get outDir from configuration");
			}
			long sizeOfElem = -1;
			boolean performInMemoryJoin = true;
			ArrayList<StringElem> inMemoryList = new ArrayList<StringElem>();
			
			log.debug("BR start inmemory collection");
			log.debug("BR CJK: " + key.toString());
			Iterator<StringElem> valItr = values.iterator();
			int counter = 0;
			while (valItr.hasNext())
			{
				StringElem ptrElem = valItr.next();
				StringElem currElem = new StringElem(ptrElem.toString());
				currElem.setPartitionID(ptrElem.getPartitionID());
				currElem.setWindowID(ptrElem.getWindowID());
				currElem.setPrevPartition(ptrElem.getPrevPartition());
				
				
				if(sizeOfElem==-1)
					sizeOfElem = currElem.getSizeInBytes();
				
				inMemoryList.add(counter, currElem);
				counter++;
				
				//if we have reached the limit of memory and there is more to do, break and don't perform inMemoryJoin
				if((
						memory <= ((inMemoryList.size()) * sizeOfElem)) && valItr.hasNext())
				{
					log.debug("BR we are not performing in memory join");
					performInMemoryJoin = false;
					break;					
				}
			}
			
			if(performInMemoryJoin)
			{
				log.debug("BR performInMemoryJoin - Base");
				//for all elements in inMemoryList, 
				//check if they can be joined to all others
				//if they can, write output to final output
				inMemoryList.trimToSize();
				baseQuickJoin(inMemoryList.subList(0, inMemoryList.size()), eps, context, key.getWindowID());
				
				
			}//end in memory join
			else//perform out of memory operations
			{
				FileSystem fs;															//create variable for file system access
				Path outfile;
				Path itrOutDir;
				
				FSDataOutputStream[] out = new FSDataOutputStream[numFilesToGen];
				
				log.debug("BR performOutOfMemoryOperation");
				
				counter = 0;
				
				if(key.getWindowID()==-1)											//it's a normal partition
				{	
					try {
						fs = FileSystem.get(context.getConfiguration());					//try to set up a output stream to the HDFS
						
						itrOutDir = new Path(outDir + "intermediate_output_" + itr + "_" + key.getPartitionID());
						fs.mkdirs(itrOutDir);
						for(int i = 0; i < numFilesToGen; i++)
						{
							outfile = new Path(itrOutDir.toString() + "/" + itr + "_" + i + ".txt");
							out[i] = fs.create(outfile);
						}
						log.debug("Output Dir: " + itrOutDir.toString());
					} catch (IOException e) {
						log.debug("BR WindowID = -1: Error creating directory/file");
						log.debug(e.getMessage());						//error: unable to create stream
						e.printStackTrace();
						return;
					}
					
					while(valItr.hasNext())
					{
						StringElem currElem = valItr.next();
						try 
						{
							out[counter % numFilesToGen].writeBytes(currElem.toString() + '\n');
							counter++;
						} catch (IOException e) {
							log.debug("BR WindowID = -1: Error writing remaining elements in que");
							log.debug(e.getMessage());						
							e.printStackTrace();
						}
					}
					
					Iterator<StringElem> inMemoryItr = inMemoryList.iterator();
					while(inMemoryItr.hasNext())
					{
						StringElem currElem = inMemoryItr.next();
						try 
						{
							out[counter % numFilesToGen].writeBytes(currElem.toString() + '\n');
							counter++;
						} catch (IOException e) {
							log.debug("BR WindowID = -1: Error writing inMemoryList elements");
							log.debug(e.getMessage());						
							e.printStackTrace();
						}
					}
					for(int i = 0; i < out.length; i++)
						try {
							out[i].close();
						} catch (IOException e) {
							log.debug("BR WindowID != -1: Error closing outfiles");
							e.printStackTrace();
						}
				}//end it's a normal partion
				else																//it's a window partition
				{
					long min = Math.min(key.getPartitionID(), key.getWindowID());
					long max = Math.max(key.getPartitionID(), key.getWindowID());
					try {
						fs = FileSystem.get(context.getConfiguration());					//try to set up a output stream to the HDFS
						itrOutDir = new Path(outDir + "intermediate_output_" + itr + "_" + min + "_" + max);
						fs.mkdirs(itrOutDir);
						for(int i = 0; i < numFilesToGen; i++)
						{
							outfile = new Path(itrOutDir.toString() + "/" + itr + "_" + i + ".txt");
							out[i] = fs.create(outfile);
						}
						log.debug("Output Dir: " + itrOutDir.toString());
					} catch (IOException e) {
						log.debug("BR WindowID != -1: Error creating directory/file");
						log.debug(e.getMessage());						
						e.printStackTrace();						//error: unable to create stream
						return;
					}
					
					while(valItr.hasNext())
					{
						StringElem currElem = valItr.next();
						try 
						{
							out[counter % numFilesToGen].writeBytes(currElem.toStringPart() + '\n');
							counter++;
						} catch (IOException e) {
							log.debug("BR WindowID != -1: Error writing remaining elements in que");
							log.debug(e.getMessage());						
							e.printStackTrace();
						}
					}
					
					Iterator<StringElem> inMemoryItr = inMemoryList.iterator();
					while(inMemoryItr.hasNext())
					{
						StringElem currElem = inMemoryItr.next();
						try 
						{
							out[counter % numFilesToGen].writeBytes(currElem.toStringPart() + '\n');
							counter++;
						} catch (IOException e) {
							log.debug("BR WindowID != -1: Error writing inMemoryList elements");
							log.debug(e.getMessage());						
							e.printStackTrace();
						}
					}
					for(int i = 0; i < out.length; i++)
						try {
							out[i].close();
						} catch (IOException e) {
							log.debug("BR WindowID != -1: Error closing outfiles");
							e.printStackTrace();
						}
				}//end it's a window partition
			}//end out of memory operations
			log.debug("Intermediate elements commited to HDFS: " + counter);
			log.debug("===================================Base Reduce End===================================");
		}//end reduce	
			
		private void baseQuickJoin(List<StringElem> objs, double eps, Context context, long code)
		{
			log.debug("Begin: baseQuickJoin size: " + objs.size());
			//log.debug("objs.size() = " + objs.size() + ": baseQuickJoin");
			if(objs.size() < constSmallNum)
			{
				//log.debug("Call Nested Loop: baseQuickJoin");
				baseNestedLoop(objs, eps, context, code);
				return;
			}
			else
			{
				int p1Index;
				int p2Index;
				
				p1Index = (int)(Math.random()*(objs.size()));
				do
				{
					p2Index = (int)(Math.random()*(objs.size()));
				} while (p1Index == p2Index);
				
				//log.debug("P1Index " + objs.get(p1Index).toString() + ": baseQuickJoin");
				//log.debug("P2Index " + objs.get(p2Index).toString() + ": baseQuickJoin");
				
				//log.debug("Call basePartition: baseQuickJoin");
				basePartition(objs, eps, context, code, objs.get(p1Index), objs.get(p2Index));	
			}
		}//end BaseQuickJoin
		
		private void basePartition(List<StringElem> objs, double eps, Context context, long code,	StringElem p1, StringElem p2) 
		{
			log.debug("Begin: basePartition size: " + objs.size());
			int r 				= 0;
			int startIndx 		= 0;
			int endIndx 		= objs.size() - 1;
			
			ArrayList<StringElem> winP1 = new ArrayList<StringElem>();
			ArrayList<StringElem> winP2 = new ArrayList<StringElem>();
			
			//double startDist 	= ( (objs.get(startIndx).getDistanceBetween(p1) * objs.get(startIndx).getDistanceBetween(p1)) - (objs.get(startIndx).getDistanceBetween(p2) * objs.get(startIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double startDist 	= (objs.get(startIndx).getDistanceBetween(p1) - objs.get(startIndx).getDistanceBetween(p2)) / 2;
			//double endDist 		= ( (objs.get(endIndx).getDistanceBetween(p1) * objs.get(endIndx).getDistanceBetween(p1)) - (objs.get(endIndx).getDistanceBetween(p2) * objs.get(endIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double endDist 	= (objs.get(endIndx).getDistanceBetween(p1) - objs.get(endIndx).getDistanceBetween(p2)) / 2;
			while(startIndx < endIndx)
			{
				
				while(endDist > r && startIndx < endIndx)
				{//closest to p2
					if(endDist <= (r + (eps + errorAdjsmt) ))
						winP1.add(objs.get(endIndx));
					endIndx--;
					//endDist 		= ( (objs.get(endIndx).getDistanceBetween(p1) * objs.get(endIndx).getDistanceBetween(p1)) - (objs.get(endIndx).getDistanceBetween(p2) * objs.get(endIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					endDist 	= (objs.get(endIndx).getDistanceBetween(p1) - objs.get(endIndx).getDistanceBetween(p2)) / 2;	
				}
				
				while(startDist <= r && startIndx < endIndx)
				{//closest to p1
					if(startDist >= (r - (eps + errorAdjsmt) ) && startIndx < endIndx)
						winP2.add(objs.get(startIndx));
					startIndx++;
					//startDist 	= ( (objs.get(startIndx).getDistanceBetween(p1) * objs.get(startIndx).getDistanceBetween(p1)) - (objs.get(startIndx).getDistanceBetween(p2) * objs.get(startIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					startDist 	= (objs.get(startIndx).getDistanceBetween(p1) - objs.get(startIndx).getDistanceBetween(p2)) / 2;
				}
				
				if(startIndx < endIndx)
				{
					if(endDist >= (r - (eps + errorAdjsmt) ))
						winP2.add(objs.get(endIndx));
					if(startDist <= (r + (eps + errorAdjsmt) ))
						winP1.add(objs.get(startIndx));
					
					//exchange positions//
					StringElem temp = objs.get(startIndx);
					objs.set(startIndx, objs.get(endIndx));
					objs.set(endIndx, temp);
					//////////////////////
					
					startIndx++;
					//startDist 	= ( (objs.get(startIndx).getDistanceBetween(p1) * objs.get(startIndx).getDistanceBetween(p1)) - (objs.get(startIndx).getDistanceBetween(p2) * objs.get(startIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					startDist 	= (objs.get(startIndx).getDistanceBetween(p1) - objs.get(startIndx).getDistanceBetween(p2)) / 2;
					
					endIndx--;
					//endDist 		= ( (objs.get(endIndx).getDistanceBetween(p1) * objs.get(endIndx).getDistanceBetween(p1)) - (objs.get(endIndx).getDistanceBetween(p2) * objs.get(endIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					endDist 	= (objs.get(endIndx).getDistanceBetween(p1) - objs.get(endIndx).getDistanceBetween(p2)) / 2;
					
				}
				
			}//end outer while loop
			
			if(startIndx == endIndx)
			{
				if(endDist > r && endDist <= (r + (eps + errorAdjsmt) ))
					winP1.add(objs.get(endIndx));
				
				if(startDist <= r && startDist >= (r - (eps + errorAdjsmt) ))
					winP2.add(objs.get(startIndx));
				
				if(endDist > r)
					endIndx--;
			}
			
			//log.debug("objs.size() = " + objs.size() + ": basePartition");
			//log.debug("endIndx = " + endIndx + ": basePartition");
			
			//log.debug("Call baseQuickJoin: basePartition");
			baseQuickJoin(objs.subList(0, endIndx + 1), eps, context, code);
			baseQuickJoin(objs.subList(endIndx + 1, objs.size()), eps, context, code);
			//log.debug("trim: basePartition");
			winP1.trimToSize();
			winP2.trimToSize();
			//log.debug("Call baseQuickJoinWin: basePartition");
			baseQuickJoinWin(winP1.subList(0, winP1.size()), winP2.subList(0, winP2.size()), eps, context, code);
			
		}//end basePartition
				
		private void baseQuickJoinWin(List<StringElem> objs1, List<StringElem> objs2, double eps, Context context, long code) 
		{
			log.debug("Begin: baseQuickJoinWin size1: " + objs1.size() + " size2: " + objs2.size());
			int totalSize = (objs1.size() + objs2.size());
			if(totalSize < constSmallNum)
			{
				//log.debug("Call Nested Loop: baseQuickJoinWin");
				baseNestedLoop(objs1, objs2, eps, context, code);
				return;
			}
			else
			{
				int p1Index;
				int p2Index;
				
				p1Index = (int)(Math.random()*(totalSize));
				do
				{
					p2Index = (int)(Math.random()*(totalSize));
				} while (p1Index == p2Index);
				
				if(objs1.size() != 0 && objs2.size() != 0)
				{
					//log.debug("Call basePartitionWin 01: baseQuickJoinWin");
					basePartitionWin(objs1, objs2, eps, context, code, (p1Index < objs1.size() ? (objs1.get(p1Index)):(objs2.get(p1Index - objs1.size()))), (p2Index < objs1.size() ? (objs1.get(p2Index)):(objs2.get(p2Index - objs1.size()))));
				}
				else if (objs1.size() == 0)
				{// objs1 is empty, use objs2
					//log.debug("Call basePartitionWin 02: baseQuickJoinWin");
					//basePartitionWin(objs1, objs2, eps, context, objs2.get(p1Index), objs2.get(p2Index));
					return;
				}
				else
				{// objs2 is empty, use objs1
					//log.debug("Call basePartitionWin 03: baseQuickJoinWin");
					//basePartitionWin(objs1, objs2, eps, context, objs1.get(p1Index), objs1.get(p2Index));
					return;
				}
			}
			
		}//end baseQuickJoinWin	
		
		private void basePartitionWin(List<StringElem> objs1, List<StringElem> objs2, double eps, Context context, long code, StringElem p1, StringElem p2) 
		{
			log.debug("Begin: basePartitionWin size1: " + objs1.size() + " size2: " + objs2.size());
			int r 				= 0;
			
			/////////////////////////////////////////////////////////////////////////////////////////////
			int o1StartIndx 		= 0;
			int o1EndIndx 		= objs1.size() - 1;
			
			ArrayList<StringElem> o1WinP1 = new ArrayList<StringElem>();
			ArrayList<StringElem> o1WinP2 = new ArrayList<StringElem>();
			
			//double o1StartDist 	= ( (objs1.get(o1StartIndx).getDistanceBetween(p1) * objs1.get(o1StartIndx).getDistanceBetween(p1)) - (objs1.get(o1StartIndx).getDistanceBetween(p2) * objs1.get(o1StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double o1StartDist 	= (objs1.get(o1StartIndx).getDistanceBetween(p1) - objs1.get(o1StartIndx).getDistanceBetween(p2)) / 2;
			//double o1EndDist 		= ( (objs1.get(o1EndIndx).getDistanceBetween(p1) * objs1.get(o1EndIndx).getDistanceBetween(p1)) - (objs1.get(o1EndIndx).getDistanceBetween(p2) * objs1.get(o1EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double o1EndDist 	= (objs1.get(o1EndIndx).getDistanceBetween(p1) - objs1.get(o1EndIndx).getDistanceBetween(p2)) / 2;
			while(o1StartIndx < o1EndIndx)
			{
				
				while(o1EndDist > r && o1StartIndx < o1EndIndx)
				{//closest to p2
					if(o1EndDist <= (r + (eps + errorAdjsmt) ))
						o1WinP1.add(objs1.get(o1EndIndx));
					o1EndIndx--;
					//o1EndDist 		= ( (objs1.get(o1EndIndx).getDistanceBetween(p1) * objs1.get(o1EndIndx).getDistanceBetween(p1)) - (objs1.get(o1EndIndx).getDistanceBetween(p2) * objs1.get(o1EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o1EndDist 	= (objs1.get(o1EndIndx).getDistanceBetween(p1) - objs1.get(o1EndIndx).getDistanceBetween(p2)) / 2;
				}
				
				while(o1StartDist <= r && o1StartIndx < o1EndIndx)
				{//closest to p1
					if(o1StartDist >= (r - (eps + errorAdjsmt) ))
						o1WinP2.add(objs1.get(o1StartIndx));
					o1StartIndx++;
					//o1StartDist 	= ( (objs1.get(o1StartIndx).getDistanceBetween(p1) * objs1.get(o1StartIndx).getDistanceBetween(p1)) - (objs1.get(o1StartIndx).getDistanceBetween(p2) * objs1.get(o1StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o1StartDist 	= (objs1.get(o1StartIndx).getDistanceBetween(p1) - objs1.get(o1StartIndx).getDistanceBetween(p2)) / 2;
				}
				
				if(o1StartIndx < o1EndIndx)
				{
					if(o1EndDist >= (r - (eps + errorAdjsmt) ))
						o1WinP2.add(objs1.get(o1EndIndx));
					if(o1StartDist <= (r + (eps + errorAdjsmt) ))
						o1WinP1.add(objs1.get(o1StartIndx));
					
					//exchange positions//
					StringElem temp = objs1.get(o1StartIndx);
					objs1.set(o1StartIndx, objs1.get(o1EndIndx));
					objs1.set(o1EndIndx, temp);
					//////////////////////
					
					o1StartIndx++;
					//o1StartDist 	= ( (objs1.get(o1StartIndx).getDistanceBetween(p1) * objs1.get(o1StartIndx).getDistanceBetween(p1)) - (objs1.get(o1StartIndx).getDistanceBetween(p2) * objs1.get(o1StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o1StartDist 	= (objs1.get(o1StartIndx).getDistanceBetween(p1) - objs1.get(o1StartIndx).getDistanceBetween(p2)) / 2;
					
					o1EndIndx--;
					//o1EndDist 		= ( (objs1.get(o1EndIndx).getDistanceBetween(p1) * objs1.get(o1EndIndx).getDistanceBetween(p1)) - (objs1.get(o1EndIndx).getDistanceBetween(p2) * objs1.get(o1EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o1EndDist 	= (objs1.get(o1EndIndx).getDistanceBetween(p1) - objs1.get(o1EndIndx).getDistanceBetween(p2)) / 2;
					
				}
				
			}//end outer while loop
			
			if(o1StartIndx == o1EndIndx)
			{
				if(o1EndDist > r && o1EndDist <= (r + (eps + errorAdjsmt) ))
					o1WinP1.add(objs1.get(o1EndIndx));
				
				if(o1StartDist <= r && o1StartDist >= (r - (eps + errorAdjsmt) ))
					o1WinP2.add(objs1.get(o1StartIndx));
				
				if(o1EndDist > r)
					o1EndIndx--;
			}
			///////////////////////////////////////////////////////
			
			///////////////////////////////////////////////////////
			int o2StartIndx 		= 0;
			int o2EndIndx 		= objs2.size() - 1;
			
			ArrayList<StringElem> o2WinP1 = new ArrayList<StringElem>();
			ArrayList<StringElem> o2WinP2 = new ArrayList<StringElem>();
			
			//double o2StartDist 	= ( (objs2.get(o2StartIndx).getDistanceBetween(p1) * objs2.get(o2StartIndx).getDistanceBetween(p1)) - (objs2.get(o2StartIndx).getDistanceBetween(p2) * objs2.get(o2StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double o2StartDist 	= (objs2.get(o2StartIndx).getDistanceBetween(p1) - objs2.get(o2StartIndx).getDistanceBetween(p2)) / 2;
			//double o2EndDist 	= ( (objs2.get(o2EndIndx).getDistanceBetween(p1) * objs2.get(o2EndIndx).getDistanceBetween(p1)) - (objs2.get(o2EndIndx).getDistanceBetween(p2) * objs2.get(o2EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double o2EndDist 	= (objs2.get(o2EndIndx).getDistanceBetween(p1) - objs2.get(o2EndIndx).getDistanceBetween(p2)) / 2;
			while(o2StartIndx < o2EndIndx)
			{
				
				while(o2EndDist > r && o2StartIndx < o2EndIndx)
				{//closest to p2
					if(o2EndDist <= (r + (eps + errorAdjsmt) ))
						o2WinP1.add(objs2.get(o2EndIndx));
					o2EndIndx--;
					//o2EndDist 	= ( (objs2.get(o2EndIndx).getDistanceBetween(p1) * objs2.get(o2EndIndx).getDistanceBetween(p1)) - (objs2.get(o2EndIndx).getDistanceBetween(p2) * objs2.get(o2EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o2EndDist 	= (objs2.get(o2EndIndx).getDistanceBetween(p1) - objs2.get(o2EndIndx).getDistanceBetween(p2)) / 2;
				}
				
				while(o2StartDist <= r && o2StartIndx < o2EndIndx)
				{//closest to p1
					if(o2StartDist >= (r - (eps + errorAdjsmt) ))
						o2WinP2.add(objs2.get(o2StartIndx));
					o2StartIndx++;
					//o2StartDist 	= ( (objs2.get(o2StartIndx).getDistanceBetween(p1) * objs2.get(o2StartIndx).getDistanceBetween(p1)) - (objs2.get(o2StartIndx).getDistanceBetween(p2) * objs2.get(o2StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o2StartDist 	= (objs2.get(o2StartIndx).getDistanceBetween(p1) - objs2.get(o2StartIndx).getDistanceBetween(p2)) / 2;
				}
				
				if(o2StartIndx < o2EndIndx)
				{
					if(o2EndDist >= (r - (eps + errorAdjsmt) ))
						o2WinP2.add(objs2.get(o2EndIndx));
					if(o2StartDist <= (r + (eps + errorAdjsmt) ))
						o2WinP1.add(objs2.get(o2StartIndx));
					
					//exchange positions//
					StringElem temp = objs2.get(o2StartIndx);
					objs2.set(o2StartIndx, objs2.get(o2EndIndx));
					objs2.set(o2EndIndx, temp);
					//////////////////////
					
					o2StartIndx++;
					//o2StartDist 	= ( (objs2.get(o2StartIndx).getDistanceBetween(p1) * objs2.get(o2StartIndx).getDistanceBetween(p1)) - (objs2.get(o2StartIndx).getDistanceBetween(p2) * objs2.get(o2StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o2StartDist 	= (objs2.get(o2StartIndx).getDistanceBetween(p1) - objs2.get(o2StartIndx).getDistanceBetween(p2)) / 2;
					
					o2EndIndx--;
					//o2EndDist 	= ( (objs2.get(o2EndIndx).getDistanceBetween(p1) * objs2.get(o2EndIndx).getDistanceBetween(p1)) - (objs2.get(o2EndIndx).getDistanceBetween(p2) * objs2.get(o2EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o2EndDist 	= (objs2.get(o2EndIndx).getDistanceBetween(p1) - objs2.get(o2EndIndx).getDistanceBetween(p2)) / 2;
					
				}
				
			}//end outer while loop
			
			if(o2StartIndx == o2EndIndx)
			{
				if(o2EndDist > r && o2EndDist <= (r + (eps + errorAdjsmt) ))
					o2WinP1.add(objs2.get(o2EndIndx));
				
				if(o2StartDist <= r && o2StartDist >= (r - (eps + errorAdjsmt) ))
					o2WinP2.add(objs2.get(o2StartIndx));
				
				if(o2EndDist > r)
					o2EndIndx--;
			}
			
			//log.debug("trim: basePartitionWin");
			o1WinP1.trimToSize();
			o1WinP2.trimToSize();
			o2WinP1.trimToSize();
			o2WinP2.trimToSize();
			
			//log.debug("Call baseQuickJoinWin: basePartitionWin");
			baseQuickJoinWin(o1WinP1.subList(0, o1WinP1.size()), o2WinP2.subList(0, o2WinP2.size()), eps, context, code);
			baseQuickJoinWin(o1WinP2.subList(0, o1WinP2.size()), o2WinP1.subList(0, o2WinP1.size()), eps, context, code);
			baseQuickJoinWin(objs1.subList(0, o1EndIndx + 1), objs2.subList(0, o2EndIndx + 1), eps, context, code);
			baseQuickJoinWin(objs1.subList(o1EndIndx + 1, objs1.size()), objs2.subList(o2EndIndx + 1, objs2.size()), eps, context, code);
			
		}
		
		private void baseNestedLoop(List<StringElem> objs1, List<StringElem> objs2, double eps, Context context, long code) 
		{
			//log.debug("BR performInMemoryJoin - Window");
			//Iterator<StringElem> outer = inMemoryList.iterator();				//nested loop join for in memory join (n^2): NEEDS IMPROVEMENT
			//while (outer.hasNext())
			for(int i = 0; i < objs1.size(); i++)
			{
				StringElem outerElem = objs1.get(i);
				//log.debug("BR SimJoin (outer): " + outerElem.toString());
				for(int j = 0; j < objs2.size(); j++)
				{
					//StringElem innerElem = inner.next();
					StringElem innerElem = objs2.get(j);
					//log.debug("BR SimJoin (inner): " + innerElem.toString());
					if((innerElem.getKey()%2) != (outerElem.getKey()%2))
					{
						if(code == -1)
						{
						
							if (innerElem.getKey() != outerElem.getKey())
							{
								double diffDist = outerElem.getDistanceBetween(innerElem);
								//log.debug("BR Checking: Elem: " + outerElem.toString() + " Against: " + innerElem.toString() + " With Distance: " + diffDist);
								if(diffDist <= eps)										//checks if the difference in distance is less than eps
								{														//may be modified for more detailed output?????
									Text t1 = new Text(outerElem.toString());			//set up output for joined pair
									Text t2 = new Text(innerElem.toString());
									//log.debug("BR we found a match: " + t1.toString() + " : " + t2.toString());
									//log.debug("BR we found a match: " + t2.toString() + " : " + t1.toString());
									try {
										context.write(t1, t2);
										context.write(t2, t1);	
									} catch (IOException e) {
										log.error("BR Error writing to context.");
									} catch (InterruptedException e) {
										log.error("BR Error writing to context.");
									}							
															
								}
							}
						}//end it's not a window
						else
						{
							if(outerElem.partitionID != innerElem.partitionID)
							{
								double diffDist = outerElem.getDistanceBetween(innerElem);
								//log.debug("BR Checking: Elem: " + outerElem.toString() + " Against: " + innerElem.toString() + " With Distance: " + diffDist);
								if(diffDist <= eps)										//checks if the difference in distance is less than eps
								{														//may be modified for more detailed output?????
									Text t1 = new Text(outerElem.toString());			//set up output for joined pair
									Text t2 = new Text(innerElem.toString());
									//log.debug("BR we found a match: " + t1.toString() + " : " + t2.toString());
									//log.debug("BR we found a match: " + t2.toString() + " : " + t1.toString());
									try {
										context.write(t1, t2);				//write to the output both pairs
										context.write(t2, t1);
									} catch (IOException e) {
										System.out.println("Error writing to context.");
									} catch (InterruptedException e) {
										System.out.println("Error writing to context.");
									}	
								}
							}
						}//end it is a window
					}
					
				}//end inner loop
			}//end outer loop
			
		}
		
		private void baseNestedLoop(List<StringElem> objs, double eps, Context context, long code) 
		{
			for(int i = 0; i <objs.size(); i++)
			{
				StringElem outerElem = objs.get(i);
				//log.debug("BR SimJoin (outer): " + outerElem.toString());
				for(int j = 0; j < objs.size(); j++)
				{
					StringElem innerElem = objs.get(j);
					//log.debug("BR SimJoin (inner): " + innerElem.toString());
					if((innerElem.getKey()%2) != (outerElem.getKey()%2))
					{
						if(code == -1)
						{
						
							if (innerElem.getKey() != outerElem.getKey())
							{
								double diffDist = outerElem.getDistanceBetween(innerElem);
								//log.debug("BR Checking: Elem: " + outerElem.toString() + " Against: " + innerElem.toString() + " With Distance: " + diffDist);
								if(diffDist <= eps)										//checks if the difference in distance is less than eps
								{														//may be modified for more detailed output?????
									Text t1 = new Text(outerElem.toString());			//set up output for joined pair
									Text t2 = new Text(innerElem.toString());
									//log.debug("BR we found a match: " + t1.toString() + " : " + t2.toString());
									try {
										context.write(t1, t2);				
									} catch (IOException e) {
										log.error("BR Error writing to context.");
									} catch (InterruptedException e) {
										log.error("BR Error writing to context.");
									}							
															
								}
							}
						}//end it's not a window
						else
						{
							if(outerElem.partitionID != innerElem.partitionID)
							{
								double diffDist = outerElem.getDistanceBetween(innerElem);
								//log.debug("BR Checking: Elem: " + outerElem.toString() + " Against: " + innerElem.toString() + " With Distance: " + diffDist);
								if(diffDist <= eps)										//checks if the difference in distance is less than eps
								{														//may be modified for more detailed output?????
									Text t1 = new Text(outerElem.toString());			//set up output for joined pair
									Text t2 = new Text(innerElem.toString());
									//log.debug("BR we found a match: " + t1.toString() + " : " + t2.toString());
									try {
										context.write(t1, t2);				//write to the output both pairs
									} catch (IOException e) {
										System.out.println("Error writing to context.");
									} catch (InterruptedException e) {
										System.out.println("Error writing to context.");
									}	
								}
							}
						}//end it is a window
					}
					
				}//end inner loop
			}//end outter loop
			
		}//end baseNestedLoop
		
	}//end CloudJoinBaseReducer
	
	public static class CloudJoinWindowReducer extends Reducer<CloudJoinKey, StringElem, Text, Text>
	{

		public void reduce(CloudJoinKey key, Iterable<StringElem> values, Context context)
		{
			//log.setLevel(Level.OFF);
			log.debug("===================================Window Reduce Begin===================================");
			String s;
			s = context.getConfiguration().get("eps", "");
			if("".equals(s))
			{
				System.out.println("Error: Unable to get eps from configuration");
			}
			double eps = Double.parseDouble(s);
			s = context.getConfiguration().get("memory", "");
			if("".equals(s))
			{
				System.out.println("Error: Unable to get memory from configuration");
			}
			long memory = Long.parseLong(s);
			s = context.getConfiguration().get("itr", "");
			if("".equals(s))
			{
				System.out.println("Error: Unable to get itr from configuration");
			}
			long itr = Long.parseLong(s);
			String outDir = context.getConfiguration().get("outDir", "");
			if("".equals(outDir))
			{
				System.out.println("Error: Unable to get outDir from configuration");
			}
			long sizeOfElem = -1;
			boolean performInMemoryJoin = true;
			ArrayList<StringElem> inMemoryList = new ArrayList<StringElem>();
			
			
			log.debug("WR start inmemory collection");
			//log.debug("WR CJK: " + key.toString());
			Iterator<StringElem> valItr = values.iterator();
			int counter = 0;
			while (valItr.hasNext())
			{
				StringElem ptrElem = valItr.next();
				StringElem currElem = new StringElem(ptrElem.toString());
				currElem.setPartitionID(ptrElem.getPartitionID());
				currElem.setWindowID(ptrElem.getWindowID());
				currElem.setPrevPartition(ptrElem.getPrevPartition());
				
				//log.debug("BR collecting: " + currElem.toString());
				
				if(sizeOfElem==-1)//could we set this outside the loop?
					sizeOfElem = currElem.getSizeInBytes();
				
				inMemoryList.add(counter, currElem);
				counter++;
				//log.debug("R collecting(added?): " + inMemoryList.get(inMemoryList.size() - 1).toString());
				
				//if we have reached the limit of memory and there is more to do, break and don't perform inMemoryJoin
				if((
						memory <= ((inMemoryList.size()) * sizeOfElem)) && valItr.hasNext())
				{
					log.debug("WR we are not performing in memory join");
					performInMemoryJoin = false;
					break;					
				}
			}
			
			log.debug("WR end inmemory collection");
			//log.debug("WR size of inMemoryList: " + inMemoryList.size());
			
			if(performInMemoryJoin)
			{
				//for all elements in inMemoryList, 
				//check if they can be joined to all others
				//if they can, write output to final output
				log.debug("WR performInMemoryJoin - Base");
				inMemoryList.trimToSize();
				windowQuickJoin(inMemoryList.subList(0, inMemoryList.size()), eps, context, key.getWindowID());
				
			}
			else//perform out of memory operations
			{
				FileSystem fs;															//create variable for file system access
				Path outfile;
				Path itrOutDir;
				FSDataOutputStream[] out = new FSDataOutputStream[numFilesToGen];
				counter = 0;
				log.debug("WR performOutOfMemoryOperation");
				if(key.getWindowID()==-1)											//it's a normal partition
				{	
					try {
						fs = FileSystem.get(context.getConfiguration());					//try to set up a output stream to the HDFS
						itrOutDir = new Path(outDir + "intermediate_output_" + itr + "_" + context.getConfiguration().get("V") + "_" + context.getConfiguration().get("W") + "_" + key.partitionID);
						fs.mkdirs(itrOutDir);
						
						for(int i = 0; i < numFilesToGen; i++)
						{
							outfile = new Path(itrOutDir.toString() + "/" + itr + "_" + i + ".txt");
							out[i] = fs.create(outfile);
						}
						log.debug("Output Dir: " + itrOutDir.toString());
					} catch (IOException e) {
						log.debug("WR WindowID = -1: Error creating directory/file");
						log.debug(e.getMessage());						//error: unable to create stream
						e.printStackTrace();						//error: unable to create stream
						return;
					}
					
					while(valItr.hasNext())
					{
						StringElem currElem = valItr.next();
						try 
						{
							out[counter % numFilesToGen].writeBytes(currElem.toStringPrev() + '\n');
							counter++;
						} catch (IOException e) {
							log.debug("WR WindowID = -1: Error writing remaining elements in que");
							log.debug(e.getMessage());						
							e.printStackTrace();
						}
					}
					
					Iterator<StringElem> inMemoryItr = inMemoryList.iterator();
					while(inMemoryItr.hasNext())
					{
						StringElem currElem = inMemoryItr.next();
						try 
						{
							out[counter % numFilesToGen].writeBytes(currElem.toStringPrev() + '\n');
							counter++;
						} catch (IOException e) {
							log.debug("WR WindowID = -1: Error writing inMemoryList elements");
							log.debug(e.getMessage());						
							e.printStackTrace();
						}
					}
					for(int i = 0; i < out.length; i++)
						try {
							out[i].close();
						} catch (IOException e) {
							log.debug("BR WindowID != -1: Error closing outfiles");
							e.printStackTrace();
						}
				}
				else																//it's a window partition
				{
					try {
						fs = FileSystem.get(context.getConfiguration());					//try to set up a output stream to the HDFS
						itrOutDir = new Path(outDir + "intermediate_output_" + itr + "_" + key.getPartitionID() + "_" + key.getWindowID() + "_" + key.getPrevItrPartion());
						fs.mkdirs(itrOutDir);
						
						for(int i = 0; i < numFilesToGen; i++)
						{
							outfile = new Path(itrOutDir.toString() + "/" + itr + "_" + i + ".txt");
							out[i] = fs.create(outfile);
						}
						log.debug("Output Dir: " + itrOutDir.toString());
						
					} catch (IOException e) {
						log.debug("WR WindowID != -1: Error creating directory/file");
						log.debug(e.getMessage());						//error: unable to create stream
						e.printStackTrace();						//error: unable to create stream
						return;
					}
					
					while(valItr.hasNext())
					{
						StringElem currElem = valItr.next();
						try 
						{
							out[counter % numFilesToGen].writeBytes(currElem.toStringPart() + '\n');
							counter++;
						} catch (IOException e) {
							log.debug("WR WindowID != -1: Error writing remaining elements in que");
							log.debug(e.getMessage());						
							e.printStackTrace();
						}
					}
					
					Iterator<StringElem> inMemoryItr = inMemoryList.iterator();
					while(inMemoryItr.hasNext())
					{
						StringElem currElem = inMemoryItr.next();
						try 
						{
							out[counter % numFilesToGen].writeBytes(currElem.toStringPart() + '\n');
							counter++;
						} catch (IOException e) {
							log.debug("WR WindowID != -1: Error writing inMemoryList elements");
							log.debug(e.getMessage());						
							e.printStackTrace();
						}
					}
					for(int i = 0; i < out.length; i++)
						try {
							out[i].close();
						} catch (IOException e) {
							log.debug("BR WindowID != -1: Error closing outfiles");
							e.printStackTrace();
						}
				}
				log.debug("Intermediate elements commited to HDFS: " + counter);
			}
			
			log.debug("===================================Window Reduce End===================================");
		}//end reduce	
		
		private void windowQuickJoin(List<StringElem> objs, double eps, Context context, long code)
		{
			log.debug("Begin: windowQuickJoin size1: " + objs.size());
			//log.debug("objs.size() = " + objs.size() + ": windowQuickJoin");
			if(objs.size() < constSmallNum)
			{
				//log.debug("Call Nested Loop: windowQuickJoin");
				windowNestedLoop(objs, eps, context, code);
				return;
			}
			else
			{
				int p1Index;
				int p2Index;
				
				p1Index = (int)(Math.random()*(objs.size()));
				do
				{
					p2Index = (int)(Math.random()*(objs.size()));
				} while (p1Index == p2Index);
				
				//log.debug("P1Index " + objs.get(p1Index).toString() + ": windowQuickJoin");
				//log.debug("P2Index " + objs.get(p2Index).toString() + ": windowQuickJoin");
				
				//log.debug("Call basePartition: windowQuickJoin");
				windowPartition(objs, eps, context, code, objs.get(p1Index), objs.get(p2Index));	
			}
		}//end BaseQuickJoin
		
		private void windowPartition(List<StringElem> objs, double eps, Context context, long code,	StringElem p1, StringElem p2) 
		{
			log.debug("Begin: windowPartition size1: " + objs.size());
			int r 				= 0;
			int startIndx 		= 0;
			int endIndx 		= objs.size() - 1;
			
			ArrayList<StringElem> winP1 = new ArrayList<StringElem>();
			ArrayList<StringElem> winP2 = new ArrayList<StringElem>();
			
			//double startDist 	= ( (objs.get(startIndx).getDistanceBetween(p1) * objs.get(startIndx).getDistanceBetween(p1)) - (objs.get(startIndx).getDistanceBetween(p2) * objs.get(startIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double startDist 	= (objs.get(startIndx).getDistanceBetween(p1) - objs.get(startIndx).getDistanceBetween(p2)) / 2;
			//double endDist 		= ( (objs.get(endIndx).getDistanceBetween(p1) * objs.get(endIndx).getDistanceBetween(p1)) - (objs.get(endIndx).getDistanceBetween(p2) * objs.get(endIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double endDist 	= (objs.get(endIndx).getDistanceBetween(p1) - objs.get(endIndx).getDistanceBetween(p2)) / 2;
			while(startIndx < endIndx)
			{
				
				while(endDist > r && startIndx < endIndx)
				{//closest to p2
					if(endDist <= (r + (eps + errorAdjsmt) ))
						winP1.add(objs.get(endIndx));
					endIndx--;
					//endDist 		= ( (objs.get(endIndx).getDistanceBetween(p1) * objs.get(endIndx).getDistanceBetween(p1)) - (objs.get(endIndx).getDistanceBetween(p2) * objs.get(endIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					endDist 	= (objs.get(endIndx).getDistanceBetween(p1) - objs.get(endIndx).getDistanceBetween(p2)) / 2;	
				}
				
				while(startDist <= r && startIndx < endIndx)
				{//closest to p1
					if(startDist >= (r - (eps + errorAdjsmt) ) && startIndx < endIndx)
						winP2.add(objs.get(startIndx));
					startIndx++;
					//startDist 	= ( (objs.get(startIndx).getDistanceBetween(p1) * objs.get(startIndx).getDistanceBetween(p1)) - (objs.get(startIndx).getDistanceBetween(p2) * objs.get(startIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					startDist 	= (objs.get(startIndx).getDistanceBetween(p1) - objs.get(startIndx).getDistanceBetween(p2)) / 2;
				}
				
				if(startIndx < endIndx)
				{
					if(endDist >= (r - (eps + errorAdjsmt) ))
						winP2.add(objs.get(endIndx));
					if(startDist <= (r + (eps + errorAdjsmt) ))
						winP1.add(objs.get(startIndx));
					
					//exchange positions//
					StringElem temp = objs.get(startIndx);
					objs.set(startIndx, objs.get(endIndx));
					objs.set(endIndx, temp);
					//////////////////////
					
					startIndx++;
					//startDist 	= ( (objs.get(startIndx).getDistanceBetween(p1) * objs.get(startIndx).getDistanceBetween(p1)) - (objs.get(startIndx).getDistanceBetween(p2) * objs.get(startIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					startDist 	= (objs.get(startIndx).getDistanceBetween(p1) - objs.get(startIndx).getDistanceBetween(p2)) / 2;
					
					endIndx--;
					//endDist 		= ( (objs.get(endIndx).getDistanceBetween(p1) * objs.get(endIndx).getDistanceBetween(p1)) - (objs.get(endIndx).getDistanceBetween(p2) * objs.get(endIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					endDist 	= (objs.get(endIndx).getDistanceBetween(p1) - objs.get(endIndx).getDistanceBetween(p2)) / 2;
					
				}
				
			}//end outer while loop
			
			if(startIndx == endIndx)
			{
				if(endDist > r && endDist <= (r + (eps + errorAdjsmt) ))
					winP1.add(objs.get(endIndx));
				
				if(startDist <= r && startDist >= (r - (eps + errorAdjsmt) ))
					winP2.add(objs.get(startIndx));
				
				if(endDist > r)
					endIndx--;
			}
			
			//log.debug("objs.size() = " + objs.size() + ": windowPartition");
			//log.debug("endIndx = " + endIndx + ": windowPartition");
			
			//log.debug("Call windowQuickJoin: windowPartition");
			windowQuickJoin(objs.subList(0, endIndx + 1), eps, context, code);
			windowQuickJoin(objs.subList(endIndx + 1, objs.size()), eps, context, code);
			//log.debug("trim: windowPartition");
			winP1.trimToSize();
			winP2.trimToSize();
			//log.debug("Call windoweQuickJoinWin: windowPartition");
			windowQuickJoinWin(winP1.subList(0, winP1.size()), winP2.subList(0, winP2.size()), eps, context, code);
			
		}//end windowPartition
			
		private void windowQuickJoinWin(List<StringElem> objs1, List<StringElem> objs2, double eps, Context context, long code) 
		{
			log.debug("Begin: windowQuickJoinWin size1: " + objs1.size() + " size2: " + objs2.size());
			int totalSize = (objs1.size() + objs2.size());
			if(totalSize < constSmallNum)
			{
				//log.debug("Call Nested Loop: windowQuickJoinWin");
				windowNestedLoop(objs1, objs2, eps, context, code);
				return;
			}
			else
			{
				int p1Index;
				int p2Index;
				
				p1Index = (int)(Math.random()*(totalSize));
				do
				{
					p2Index = (int)(Math.random()*(totalSize));
				} while (p1Index == p2Index);
				
				if(objs1.size() != 0 && objs2.size() != 0)
				{
					//log.debug("Call windowPartitionWin 01: windowQuickJoinWin");
					windowPartitionWin(objs1, objs2, eps, context, code, (p1Index < objs1.size() ? (objs1.get(p1Index)):(objs2.get(p1Index - objs1.size()))), (p2Index < objs1.size() ? (objs1.get(p2Index)):(objs2.get(p2Index - objs1.size()))));
				}
				else if (objs1.size() == 0)
				{// objs1 is empty, use objs2
					//log.debug("Call windowPartitionWin 02: windowQuickJoinWin");
					//basePartitionWin(objs1, objs2, eps, context, objs2.get(p1Index), objs2.get(p2Index));
					return;
				}
				else
				{// objs2 is empty, use objs1
					//log.debug("Call windowPartitionWin 03: windowQuickJoinWin");
					//basePartitionWin(objs1, objs2, eps, context, objs1.get(p1Index), objs1.get(p2Index));
					return;
				}
			}
			
		}
		
		private void windowPartitionWin(List<StringElem> objs1, List<StringElem> objs2, double eps, Context context, long code, StringElem p1, StringElem p2) 
		{
			log.debug("Begin: windowPartitionWin size1: " + objs1.size() + " size2: " + objs2.size());
			int r 				= 0;
			
			/////////////////////////////////////////////////////////////////////////////////////////////
			int o1StartIndx 		= 0;
			int o1EndIndx 		= objs1.size() - 1;
			
			ArrayList<StringElem> o1WinP1 = new ArrayList<StringElem>();
			ArrayList<StringElem> o1WinP2 = new ArrayList<StringElem>();
			
			//double o1StartDist 	= ( (objs1.get(o1StartIndx).getDistanceBetween(p1) * objs1.get(o1StartIndx).getDistanceBetween(p1)) - (objs1.get(o1StartIndx).getDistanceBetween(p2) * objs1.get(o1StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double o1StartDist 	= (objs1.get(o1StartIndx).getDistanceBetween(p1) - objs1.get(o1StartIndx).getDistanceBetween(p2)) / 2;
			//double o1EndDist 		= ( (objs1.get(o1EndIndx).getDistanceBetween(p1) * objs1.get(o1EndIndx).getDistanceBetween(p1)) - (objs1.get(o1EndIndx).getDistanceBetween(p2) * objs1.get(o1EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double o1EndDist 	= (objs1.get(o1EndIndx).getDistanceBetween(p1) - objs1.get(o1EndIndx).getDistanceBetween(p2)) / 2;
			while(o1StartIndx < o1EndIndx)
			{
				
				while(o1EndDist > r && o1StartIndx < o1EndIndx)
				{//closest to p2
					if(o1EndDist <= (r + (eps + errorAdjsmt) ))
						o1WinP1.add(objs1.get(o1EndIndx));
					o1EndIndx--;
					//o1EndDist 		= ( (objs1.get(o1EndIndx).getDistanceBetween(p1) * objs1.get(o1EndIndx).getDistanceBetween(p1)) - (objs1.get(o1EndIndx).getDistanceBetween(p2) * objs1.get(o1EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o1EndDist 	= (objs1.get(o1EndIndx).getDistanceBetween(p1) - objs1.get(o1EndIndx).getDistanceBetween(p2)) / 2;
				}
				
				while(o1StartDist <= r && o1StartIndx < o1EndIndx)
				{//closest to p1
					if(o1StartDist >= (r - (eps + errorAdjsmt) ))
						o1WinP2.add(objs1.get(o1StartIndx));
					o1StartIndx++;
					//o1StartDist 	= ( (objs1.get(o1StartIndx).getDistanceBetween(p1) * objs1.get(o1StartIndx).getDistanceBetween(p1)) - (objs1.get(o1StartIndx).getDistanceBetween(p2) * objs1.get(o1StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o1StartDist 	= (objs1.get(o1StartIndx).getDistanceBetween(p1) - objs1.get(o1StartIndx).getDistanceBetween(p2)) / 2;
				}
				
				if(o1StartIndx < o1EndIndx)
				{
					if(o1EndDist >= (r - (eps + errorAdjsmt) ))
						o1WinP2.add(objs1.get(o1EndIndx));
					if(o1StartDist <= (r + (eps + errorAdjsmt) ))
						o1WinP1.add(objs1.get(o1StartIndx));
					
					//exchange positions//
					StringElem temp = objs1.get(o1StartIndx);
					objs1.set(o1StartIndx, objs1.get(o1EndIndx));
					objs1.set(o1EndIndx, temp);
					//////////////////////
					
					o1StartIndx++;
					//o1StartDist 	= ( (objs1.get(o1StartIndx).getDistanceBetween(p1) * objs1.get(o1StartIndx).getDistanceBetween(p1)) - (objs1.get(o1StartIndx).getDistanceBetween(p2) * objs1.get(o1StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o1StartDist 	= (objs1.get(o1StartIndx).getDistanceBetween(p1) - objs1.get(o1StartIndx).getDistanceBetween(p2)) / 2;
					
					o1EndIndx--;
					//o1EndDist 		= ( (objs1.get(o1EndIndx).getDistanceBetween(p1) * objs1.get(o1EndIndx).getDistanceBetween(p1)) - (objs1.get(o1EndIndx).getDistanceBetween(p2) * objs1.get(o1EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o1EndDist 	= (objs1.get(o1EndIndx).getDistanceBetween(p1) - objs1.get(o1EndIndx).getDistanceBetween(p2)) / 2;
					
				}
				
			}//end outer while loop
			
			if(o1StartIndx == o1EndIndx)
			{
				if(o1EndDist > r && o1EndDist <= (r + (eps + errorAdjsmt) ))
					o1WinP1.add(objs1.get(o1EndIndx));
				
				if(o1StartDist <= r && o1StartDist >= (r - (eps + errorAdjsmt) ))
					o1WinP2.add(objs1.get(o1StartIndx));
				
				if(o1EndDist > r)
					o1EndIndx--;
			}
			///////////////////////////////////////////////////////
			
			///////////////////////////////////////////////////////
			int o2StartIndx 		= 0;
			int o2EndIndx 		= objs2.size() - 1;
			
			ArrayList<StringElem> o2WinP1 = new ArrayList<StringElem>();
			ArrayList<StringElem> o2WinP2 = new ArrayList<StringElem>();
			
			//double o2StartDist 	= ( (objs2.get(o2StartIndx).getDistanceBetween(p1) * objs2.get(o2StartIndx).getDistanceBetween(p1)) - (objs2.get(o2StartIndx).getDistanceBetween(p2) * objs2.get(o2StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double o2StartDist 	= (objs2.get(o2StartIndx).getDistanceBetween(p1) - objs2.get(o2StartIndx).getDistanceBetween(p2)) / 2;
			//double o2EndDist 	= ( (objs2.get(o2EndIndx).getDistanceBetween(p1) * objs2.get(o2EndIndx).getDistanceBetween(p1)) - (objs2.get(o2EndIndx).getDistanceBetween(p2) * objs2.get(o2EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
			double o2EndDist 	= (objs2.get(o2EndIndx).getDistanceBetween(p1) - objs2.get(o2EndIndx).getDistanceBetween(p2)) / 2;
			while(o2StartIndx < o2EndIndx)
			{
				
				while(o2EndDist > r && o2StartIndx < o2EndIndx)
				{//closest to p2
					if(o2EndDist <= (r + (eps + errorAdjsmt) ))
						o2WinP1.add(objs2.get(o2EndIndx));
					o2EndIndx--;
					//o2EndDist 	= ( (objs2.get(o2EndIndx).getDistanceBetween(p1) * objs2.get(o2EndIndx).getDistanceBetween(p1)) - (objs2.get(o2EndIndx).getDistanceBetween(p2) * objs2.get(o2EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o2EndDist 	= (objs2.get(o2EndIndx).getDistanceBetween(p1) - objs2.get(o2EndIndx).getDistanceBetween(p2)) / 2;
				}
				
				while(o2StartDist <= r && o2StartIndx < o2EndIndx)
				{//closest to p1
					if(o2StartDist >= (r - (eps + errorAdjsmt) ))
						o2WinP2.add(objs2.get(o2StartIndx));
					o2StartIndx++;
					//o2StartDist 	= ( (objs2.get(o2StartIndx).getDistanceBetween(p1) * objs2.get(o2StartIndx).getDistanceBetween(p1)) - (objs2.get(o2StartIndx).getDistanceBetween(p2) * objs2.get(o2StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o2StartDist 	= (objs2.get(o2StartIndx).getDistanceBetween(p1) - objs2.get(o2StartIndx).getDistanceBetween(p2)) / 2;
				}
				
				if(o2StartIndx < o2EndIndx)
				{
					if(o2EndDist >= (r - (eps + errorAdjsmt) ))
						o2WinP2.add(objs2.get(o2EndIndx));
					if(o2StartDist <= (r + (eps + errorAdjsmt) ))
						o2WinP1.add(objs2.get(o2StartIndx));
					
					//exchange positions//
					StringElem temp = objs2.get(o2StartIndx);
					objs2.set(o2StartIndx, objs2.get(o2EndIndx));
					objs2.set(o2EndIndx, temp);
					//////////////////////
					
					o2StartIndx++;
					//o2StartDist 	= ( (objs2.get(o2StartIndx).getDistanceBetween(p1) * objs2.get(o2StartIndx).getDistanceBetween(p1)) - (objs2.get(o2StartIndx).getDistanceBetween(p2) * objs2.get(o2StartIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o2StartDist 	= (objs2.get(o2StartIndx).getDistanceBetween(p1) - objs2.get(o2StartIndx).getDistanceBetween(p2)) / 2;
					
					o2EndIndx--;
					//o2EndDist 	= ( (objs2.get(o2EndIndx).getDistanceBetween(p1) * objs2.get(o2EndIndx).getDistanceBetween(p1)) - (objs2.get(o2EndIndx).getDistanceBetween(p2) * objs2.get(o2EndIndx).getDistanceBetween(p2)) ) / (2 * p1.getDistanceBetween(p2));
					o2EndDist 	= (objs2.get(o2EndIndx).getDistanceBetween(p1) - objs2.get(o2EndIndx).getDistanceBetween(p2)) / 2;
					
				}
				
			}//end outer while loop
			
			if(o2StartIndx == o2EndIndx)
			{
				if(o2EndDist > r && o2EndDist <= (r + (eps + errorAdjsmt) ))
					o2WinP1.add(objs2.get(o2EndIndx));
				
				if(o2StartDist <= r && o2StartDist >= (r - (eps + errorAdjsmt) ))
					o2WinP2.add(objs2.get(o2StartIndx));
				
				if(o2EndDist > r)
					o2EndIndx--;
			}
			
			//log.debug("trim: windowPartitionWin");
			o1WinP1.trimToSize();
			o1WinP2.trimToSize();
			o2WinP1.trimToSize();
			o2WinP2.trimToSize();
			
			//log.debug("Call windowQuickJoinWin:windowPartitionWin");
			windowQuickJoinWin(o1WinP1.subList(0, o1WinP1.size()), o2WinP2.subList(0, o2WinP2.size()), eps, context, code);
			windowQuickJoinWin(o1WinP2.subList(0, o1WinP2.size()), o2WinP1.subList(0, o2WinP1.size()), eps, context, code);
			windowQuickJoinWin(objs1.subList(0, o1EndIndx + 1), objs2.subList(0, o2EndIndx + 1), eps, context, code);
			windowQuickJoinWin(objs1.subList(o1EndIndx + 1, objs1.size()), objs2.subList(o2EndIndx + 1, objs2.size()), eps, context, code);
			
		}
		
		private void windowNestedLoop(List<StringElem> objs1, List<StringElem> objs2, double eps, Context context, long code) 
		{
			
			for(int i = 0; i < objs1.size(); i++)
			{
				StringElem outerElem = objs1.get(i);
				for(int j = 0; j < objs2.size(); j++)
				{
					StringElem innerElem = objs2.get(j);
					//log.debug("BR SimJoin (inner): " + innerElem.toString());
					if((innerElem.getKey()%2) != (outerElem.getKey()%2))
					{
						if(code == -1)
						{
						
							if (innerElem.getPrevPartition() != outerElem.getPrevPartition())
							{
								double diffDist = outerElem.getDistanceBetween(innerElem);
								//log.debug("BR Checking: Elem: " + outerElem.toString() + " Against: " + innerElem.toString() + " With Distance: " + diffDist);
								if(diffDist <= eps)										//checks if the difference in distance is less than eps
								{														//may be modified for more detailed output?????
									Text t1 = new Text(outerElem.toString());			//set up output for joined pair
									Text t2 = new Text(innerElem.toString());
									//log.debug("BR we found a match: " + t1.toString() + " : " + t2.toString());
									//log.debug("BR we found a match: " + t2.toString() + " : " + t1.toString());
									try {
										context.write(t1, t2);
										context.write(t2, t1);	
									} catch (IOException e) {
										log.error("BR Error writing to context.");
									} catch (InterruptedException e) {
										log.error("BR Error writing to context.");
									}							
															
								}
							}
						}//end it's not a window
						else
						{
							if (innerElem.getPrevPartition() != outerElem.getPrevPartition())
							{
								if(outerElem.partitionID != innerElem.partitionID)
								{
									double diffDist = outerElem.getDistanceBetween(innerElem);
									//log.debug("BR Checking: Elem: " + outerElem.toString() + " Against: " + innerElem.toString() + " With Distance: " + diffDist);
									if(diffDist <= eps)										//checks if the difference in distance is less than eps
									{														//may be modified for more detailed output?????
										Text t1 = new Text(outerElem.toString());			//set up output for joined pair
										Text t2 = new Text(innerElem.toString());
										//log.debug("BR we found a match: " + t1.toString() + " : " + t2.toString());
										//log.debug("BR we found a match: " + t2.toString() + " : " + t1.toString());
										try {
											context.write(t1, t2);				//write to the output both pairs
											context.write(t2, t1);
										} catch (IOException e) {
											System.out.println("Error writing to context.");
										} catch (InterruptedException e) {
											System.out.println("Error writing to context.");
										}	
									}
								}
							}
						}//end it is a window
					}
					
				}//end inner loop
			}//end outer loop
			
		}
		
		private void windowNestedLoop(List<StringElem> objs, double eps, Context context, long code) 
		{
			for(int i = 0; i <objs.size(); i++)
			{
				StringElem outerElem = objs.get(i);
				//log.debug("BR SimJoin (outer): " + outerElem.toString());
				for(int j = 0; j < objs.size(); j++)
				{
					StringElem innerElem = objs.get(j);
					//log.debug("BR SimJoin (inner): " + innerElem.toString());
					if((innerElem.getKey()%2) != (outerElem.getKey()%2))
					{
						if(code == -1)
						{
						
							if (innerElem.getPrevPartition() != outerElem.getPrevPartition())
							{
								double diffDist = outerElem.getDistanceBetween(innerElem);
								//log.debug("BR Checking: Elem: " + outerElem.toString() + " Against: " + innerElem.toString() + " With Distance: " + diffDist);
								if(diffDist <= eps)										//checks if the difference in distance is less than eps
								{														//may be modified for more detailed output?????
									Text t1 = new Text(outerElem.toString());			//set up output for joined pair
									Text t2 = new Text(innerElem.toString());
									//log.debug("BR we found a match: " + t1.toString() + " : " + t2.toString());
									try {
										context.write(t1, t2);				
									} catch (IOException e) {
										log.error("BR Error writing to context.");
									} catch (InterruptedException e) {
										log.error("BR Error writing to context.");
									}							
															
								}
							}
						}//end it's not a window
						else
						{
							if (innerElem.getPrevPartition() != outerElem.getPrevPartition())
							{
								if(outerElem.partitionID != innerElem.partitionID)
								{
									double diffDist = outerElem.getDistanceBetween(innerElem);
									//log.debug("BR Checking: Elem: " + outerElem.toString() + " Against: " + innerElem.toString() + " With Distance: " + diffDist);
									if(diffDist <= eps)										//checks if the difference in distance is less than eps
									{														//may be modified for more detailed output?????
										Text t1 = new Text(outerElem.toString());			//set up output for joined pair
										Text t2 = new Text(innerElem.toString());
										//log.debug("BR we found a match: " + t1.toString() + " : " + t2.toString());
										try {
											context.write(t1, t2);				//write to the output both pairs
										} catch (IOException e) {
											System.out.println("Error writing to context.");
										} catch (InterruptedException e) {
											System.out.println("Error writing to context.");
										}	
									}
								}
							}
						}//end it is a window
					}
				}//end inner loop
			}//end outer loop
			
		}//end baseNestedLoop

	}
	
	public static void main(String[] args) throws Exception
	{
		int res = ToolRunner.run(new Configuration(), new CloudJoin04Ext(), args);
		System.exit(res);
	}
	
	public int run(String[] args) throws Exception 
	{
		if(args.length != 3)
		{
			System.out.println("Usage: bin/hadoop jar CloudJoin03.jar CloudJoin03 <eps> <number pivots> <number of reduces>");
			System.out.println("args length incorrect, length: " + args.length);
			return -1;
		}
		double eps;
		int numP, numReduces, numMappers;
		
		try
		{
			eps = new Double(args[0]);
			System.out.println("eps set to: " + eps);
		}
		catch(NumberFormatException e)
		{
			System.out.println("Usage: bin/hadoop jar CloudJoin03.jar CloudJoin03 <eps> <number pivots> <number of reduces>");
			System.out.println("Error: eps not a type double");
			return -1;
		}
		try
		{
			numP = new Integer(args[1]);
			System.out.println("number of pivots set to: " + numP);
		}
		catch(NumberFormatException e)
		{
			System.out.println("Usage: bin/hadoop jar CloudJoin03.jar CloudJoin03 <eps> <number pivots> <number of reduces>");
			System.out.println("Error: number of pivots not a type integer");
			return -1;
		}
		try
		{
			numReduces 	= new Integer(args[2]);
			System.out.println("number reducers set to: " + numReduces);
		}
		catch(NumberFormatException e)
		{
			System.out.println("Usage: bin/hadoop jar CloudJoin03.jar CloudJoin03 <eps> <number pivots> <number of reduces>");
			System.out.println("Error: number of reduces not a type integer");
			return -1;
		}
		
		
		String 		inputDir 		= "someInputDir/"; 
		String 		outputDir		= "someOutputDir/";
		//long 		memory 			= 131072; //128KB
		//long 		memory 			= 163840; //160KB
		//long 		memory 			= 196608; //192KB
		//long 		memory 			= 229376; //224KB
		//long 		memory 			= 262144; //256KB
		//long 		memory 			= 112 * 2730;
		long		memory			= 33554432; //32MB
		//long		memory			= 16777216; //16MB
		//long		memory			= 8388608; //8MB
		long 		itr				= 0;
		
		String 		intermediate	= outputDir + "INTERMEDIATE/";
		String 		processed		= outputDir + "PROCCESSED/";
		String 		pivDir			= outputDir + "PIVOTS/";
		
		
		long startTime = System.currentTimeMillis();
		while(true)
		{
			Configuration conf = new Configuration();
			conf.setBoolean("mapred.output.compress", false);
			conf.set("mapred.child.java.opts", "-Xmx512M");
			//conf.set("dfs.block.size", "4194304");
			//conf.set("dfs.block.size", "1048576");
			conf.set("mapred.task.timeout", "8000000");
			conf.set("itr", "" + itr);
			
			
			FileSystem fs = FileSystem.get(conf);
			Path inputPath = new Path(inputDir);
			Path pivPath = new Path(pivDir + itr + "/pivots.txt");
			boolean isBaseRound = true;   ///need to determine if it is a base round or not
			
			if(itr == 0)
			{//first iteration, set input to primary input
				
				//make intermediate directory inside outputDir
				fs.mkdirs(new Path(intermediate));
				//make processed directory inside outputDir
				fs.mkdirs(new Path(processed));
				//set input path
				inputPath = new Path(inputDir);
			}
			else
			{
				//get first input directory in INTERMEDIATE
					// use listStatus(INTERMEDIATE)
				FileStatus[] fileStat = fs.listStatus(new Path(intermediate));
				System.out.println("fileStat length: " + fileStat.length);
				if(fileStat.length >= 1)
				{
					//set jobInputDir to first directory returned
					inputPath = fileStat[0].getPath();
					System.out.println("Input Path: " + inputPath.toString());
										
				}
				else
				{
					System.out.println("No input path");
					break;
				}
				
				
				int lastIndex = inputPath.toString().lastIndexOf("intermediate_output_");
		        String subBase = inputPath.toString().substring(lastIndex + 20);
		        StringTokenizer st = new StringTokenizer(subBase, "_");
		        int tokenCount = st.countTokens();
		        if(tokenCount == 2)//if tokenCount = 2, then base partition
		        {
		        	System.out.println("CheckRound: Base Round");
		            //it's a base round, set to true and continue
		        	isBaseRound = true;
		        }
		        else if(tokenCount == 3 || tokenCount ==4)//else if tokenCount = 3 window partion
		        {
		        	System.out.println("CheckRound: Window Round");
		        	isBaseRound = false;
		            //it's a window round, set to false and continue
		            st.nextToken();
		            String v = st.nextToken();
		            String w = st.nextToken();
		            conf.set("V", v);
		            conf.set("W", w);
		            
		        }
		        else
		        {
		            System.out.println("Directory Input Error");
		            break;
		        }
				
			}
			
			
			conf.set("eps", Double.toString(eps)); 				//assign string for eps to conf
			conf.set("memory", Long.toString(memory));			//assigns the variable mem to conf
			
			Object[] pivots = getPivots(inputPath, numP);		//generate pivot points from inputPath directory
			
			FSDataOutputStream out = fs.create(pivPath);		//write pivots to file
			
			for(int i = 0; i < pivots.length; i++)
			{
				out.writeBytes(((Text)pivots[i]).toString() + '\n');
			}
			//http://hadoop.apache.org/common/docs/r0.20.2/mapred_tutorial.html#DistributedCache
			//http://hadoop.apache.org/common/docs/r0.20.2/mapred_tutorial.html#Example%3A+WordCount+v2.0
			out.close();
			System.out.println("pivPath: " + pivPath.toString());
			
			
			Job job;
			if (isBaseRound)	//if it is a Base round
			{
				//execute base round MR on jobInputDir
				System.out.println("Base Round");
				//creates job
				job = new Job(conf, "Cloud Join Base Round: Itr " + itr);
				
				job.setNumReduceTasks(numReduces);
				
				
				//add file to cache
				DistributedCache.addCacheFile(pivPath.toUri(), job.getConfiguration());
				
				job.getConfiguration().set("outDir", intermediate);
				
				
				job.setJarByClass(CloudJoin04Ext.class);
				
				//sets mapper class
				job.setMapperClass(CloudJoinBaseMapper.class);
				//sets map output key/value types
			    job.setMapOutputKeyClass(CloudJoinKey.class);
			    job.setMapOutputValueClass(StringElem.class);	
			    //sets Partitioner Class
			    job.setPartitionerClass(CloudJoinBasePartitioner.class);
			    //sets Sorting Comparitor class for job
			    job.setSortComparatorClass(CloudJoinBaseGrouper.class);
			    //Sets grouper class
			    job.setGroupingComparatorClass(CloudJoinBaseGrouper.class);
				//Set Reducer class
			    job.setReducerClass(CloudJoinBaseReducer.class);
			    // specify output types
				job.setOutputKeyClass(Text.class);
				job.setOutputValueClass(Text.class);	
				//sets output format
			    job.setInputFormatClass(TextInputFormat.class);
			    // specify input and output DIRECTORIES (not files)
				FileInputFormat.addInputPath(job, inputPath);
				FileOutputFormat.setOutputPath(job, new Path(outputDir + "/" + itr + "/"));
				
			}
			else	//else it is a window
			{
				//execute window round MR on jobInputDir
				System.out.println("Window Round");
				//set V
				//set M
				
				//creates job
				job = new Job(conf, "Cloud Join Window Round: Itr " + itr);
				job.setNumReduceTasks(numReduces);
				//add file to cache
				DistributedCache.addCacheFile(pivPath.toUri(), job.getConfiguration());
				
				job.getConfiguration().set("outDir", intermediate);

				job.setJarByClass(CloudJoin04Ext.class);

				//sets mapper class
				job.setMapperClass(CloudJoinWindowMapper.class);
				//sets map output key/value types
			    job.setMapOutputKeyClass(CloudJoinKey.class);
			    job.setMapOutputValueClass(StringElem.class);	
			    //sets Partitioner Class
			    job.setPartitionerClass(CloudJoinWindowPartitioner.class);
			    //sets Sorting Comparitor class for job
			    job.setSortComparatorClass(CloudJoinWindowGrouper.class);
			    //Sets grouper class
			    job.setGroupingComparatorClass(CloudJoinWindowGrouper.class);
				//Set Reducer class
			    job.setReducerClass(CloudJoinWindowReducer.class);
			    // specify output types
				job.setOutputKeyClass(Text.class);
				job.setOutputValueClass(Text.class);	
				//sets output format
			    job.setInputFormatClass(TextInputFormat.class);
			    // specify input and output DIRECTORIES (not files)
				FileInputFormat.addInputPath(job, inputPath);
				FileOutputFormat.setOutputPath(job, new Path(outputDir + "/" + itr + "/"));
				
			}
			
			//start and wait till it finishes
			job.waitForCompletion(true);
			
			
			if (itr > 0)
			{
				// change jobInputDir from intermediate to processed
				//	fs.rename(Path src, Path dst);
				
				String dir = inputPath.toString();
				dir = dir.replace(intermediate, processed);
				Path newPath = new Path(dir);
				fs.rename(inputPath, newPath);
			}
			
			//increment itr for next round
			itr++;
			
			
		}
		
		long endTime = System.currentTimeMillis();
		
		System.out.println("Total Run Time: " + (endTime-startTime));
		return 0;
	}
	
	Object[] getPivots(Path input, int numPivs) throws IOException
	{
		
		JobConf job = new JobConf();
		job.setInputFormat(KeyValueTextInputFormat.class);
		job.setMapOutputKeyClass(Text.class);
		org.apache.hadoop.mapred.FileInputFormat.addInputPath(job, input);
		final KeyValueTextInputFormat inf = (KeyValueTextInputFormat) job.getInputFormat();
		InputSampler.Sampler<Text, Text> sampler = new InputSampler.RandomSampler<Text, Text>(1.0, numPivs, 100);
		Object[] samples = sampler.getSample(inf, job);
		return samples;
	}
	
	
}
